/*******************************************************************************
 $Author: dmc $
 $Revision: 2688 $
 $Date: 2019-11-23 22:39:40 -0400 (Sat, 23 Nov 2019) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smt.common.lib/server_devices.js $
*******************************************************************************/

// for now... "log4js" logging needs to coexist with Adam's "log" logging...
var log4js = require( "log4js" );
var log4 = log4js.getLogger( "server-devices.js" );

log4.setLevel('ERROR');
//log4.setLevel( "TRACE" );

//log4.info("DMC test");

global.is_G2 = !!process.env.SMT_IS_G2;
global.is_R5 = !!process.env.SMT_IS_R5;
global.is_S3 = !!process.env.SMT_IS_S3;
global.is_S50 = !!process.env.SMT_IS_S50;
global.is_SOLO = !!process.env.SMT_IS_SOLO;
global.SMT_SOLO_SERIALNO = process.env.SMT_SOLO_SERIALNO;

// timers for devices to mark that they are not present
var device_present_timers = [];

function serialport_init() {
	if (is_G2 || is_R5 || is_S3) {
		if (is_G2) log4.info("We are a G2!");
		if (is_R5) log4.info("We are an R5!");
		if (is_S3) log4.info("We are an S3!");
		log4.info("starting ./serialport_monitor_G2_R5.js....");
		var G2_serial = require('./serialport_monitor_G2_R5.js');
		global.shutdown_power = G2_serial.shutdown_power;
	}
}

function faListenerInit() {
	if ( is_G2 ) return;
	// all "Server" type devices, i.e. all but a G2, listen for SOLOs
	if ( is_SOLO ) { // all "Server" type devices, i.e. all but a G2, listen for SOLOs
		log4.info("We are a SOLO!");
	}
	log4.info( "starting ./fa-server.js...." );
	require('./smt.common.lib/fa-server.js').set_stuff(global.data);
}

// start listening for UDP broadcasts and set up devices
function devices_init() {
		
	// create listeners and keep existing, associated devices
	for (var i in data.devices) {
		var device = data.devices[i];
		if (device.SOLOconnected) device.SOLOconnected = false;

		if ( device.type === "SOLO" ) {
			// for SOLO, //always purge devices on startup...
			//delete data.devices[i];
			// in principal we would have a new member device.assoc_frame_ids[...], however we probably don't need it
			if (!device.assoc_frame_id) {
				delete data.devices[i];
			}
		} else {
			if (device.assoc_frame_id) { // only keep those already associated with frames (G1 and G2)
				frame_create_listener(data.devices[i]);
			} else {
				delete data.devices[i];
			}
		}
	}
	
	// *** start listening for broadcasts ***

	// UDP broadcast of SMT devices
	// cribbed from http://stackoverflow.com/questions/6177423/send-broadcast-datagram
	var dgram = require( "dgram" );

	global.SMT_udp_port = 11223;
	global.SMT_broadcast_addr = "192.168.0.255"; // TODO: figure this out dynamically!

	global.SMT_udp_socket = dgram.createSocket( "udp4" );

	global.qndEventEmitter.on( "deviceBroadcastRecv", function ( broadcastMessage ) {
		deviceBroadcastRecv( broadcastMessage.message, broadcastMessage.ip );
	} );

	SMT_udp_socket.on( "message", function( msg, rinfo ) {

		//  log4.info("SMT dgram server got: " + msg + " from " + rinfo.address + ":" + rinfo.port);

		deviceBroadcastRecv( JSON.parse( msg ), rinfo.address );

		/*
		[2015-09-09 20:48:16.566] [INFO] server.js - SMT dgram server got: {
		"batteryvalue": 1,
		"os_hostname": "black-G2",
		"smtdevtype": "G2"
		} from 192.168.0.88:11223
		[2015-09-09 20:48:16.566] [INFO] server.js - SMT dgram server got: {
		"batteryvalue": 1,
		"os_hostname": "devboard-51-90",
		"smtdevtype": "G2"
		} from 192.168.0.101:11223
		*/
	} );

	SMT_udp_socket.on("listening", function () {
	   SMT_udp_socket.setBroadcast(true);
	//   var address = SMT_udp_socket.address();
	//   console.log("server listening " + address.address + ":" + address.port);
	 });
	
	SMT_udp_socket.bind(SMT_udp_port);
	
	/*
	var message = new Buffer("Some bytes");
	SMT_udp_socket.send(message, 0, message.length, SMT_udp_port, "0.0.0.0");
	*/

	// FSOLO faking it as a TSOLO?
	if ('string' == typeof global.SMT_SOLO_SERIALNO && global.SMT_SOLO_SERIALNO.substr(0,1) == 't') {
		init_TSOLO_emu();
	}
}


var sock_to_47server5000 = undefined;
var sock_to_47server5000_itimer = undefined;

function init_TSOLO_emu() {
	create_TSOLO_emu();
	// setInterval(create_TSOLO_emu, 20*1000); // keep retrying if failed
}

function get_carambola2_sta_sig_strength(callback/*err,sig*/) {
	// fetch signal strength
	exec( "iw dev wlan0 station dump | fgrep signal: | head -1 | cut  -f3 | cut -d' ' -f1", function(err,stdout,stderr) {
		var sig;
		if (err) {
			sig='-99';
		} else {
			sig=stdout.trim();
		}
		if ('function' == typeof callback) {
			callback(err,sig);
		}
	});
}

global.get_carambola2_sta_sig_strength = get_carambola2_sta_sig_strength;

function get_carambola2_sta_connected(callback/*is_connected*/) {
	// fetch signal strength
	exec( "iw dev wlan0 station dump", function(err,stdout,stderr) {
		var is_connected = false;
		if (err) {
		} else {
			is_connected = stdout.trim().length != 0;
		}
		if ('function' == typeof callback) {
			callback(is_connected);
		}
	});
}

global.get_carambola2_sta_connected = get_carambola2_sta_connected;

function set_TSOLO_emu_orfer() {
	clear_TSOLO_emu_orfer();
	sock_to_47server5000_itimer = setInterval( function() {
		// fetch signal strength
		get_carambola2_sta_sig_strength(function(err,sig) {
//			var msg = "ORF-"+global.SMT_SOLO_SERIALNO+"-4095,1224,4095,1233 "+sig+"\r\n"; // old format, no next_shot_no
			var msg = "ORF-"+global.SMT_SOLO_SERIALNO+"-4095,1224,4095,1233 0,"+sig+"\r\n"; // next_shot_no == 0
			// console.log("TSOLO_emu orfer()",msg.trim());
			sock_to_47server5000.write(msg);
		});
	},2000);
}

function send_orf_hw() {
	if (!sock_to_47server5000) return;

	var msg = "ORF-HW:222-FW:123\r\n"; // pretend values, but hopefully unique to us
	//console.log("TSOLO_emu send_orf_hw()",msg);
	sock_to_47server5000.write(msg);
}

function clear_TSOLO_emu_orfer() {
	if (!sock_to_47server5000_itimer) return;
	clearInterval(sock_to_47server5000_itimer);
	sock_to_47server5000_itimer = undefined;
}

function create_TSOLO_emu() {
	if (sock_to_47server5000) {
		// console.log("create_TSOLO_emu() -- connection already established");
		return;
	}
	console.log("create_TSOLO_emu() -- initiating connection");

	var net = require('net');
	sock_to_47server5000 = new net.Socket();
	sock_to_47server5000.connect(5000,"192.168.0.47",function(e,x) {
		//console.log("create_TSOLO_emu() connection established");
		set_TSOLO_emu_orfer();
	});
	sock_to_47server5000.on('error',function(errcode) {
		console.trace("create_TSOLO_emu() on 'error'",errcode);
	});
	// no need to handle 'error', since 'close' will be called anyway
	sock_to_47server5000.on('close',function() {
		console.log("create_TSOLO_emu() on 'close'");
		clear_TSOLO_emu_orfer();
		sock_to_47server5000 = undefined;
		setTimeout(create_TSOLO_emu,5000); // tail-call to retry in 5s
	});
	sock_to_47server5000.on('data',function(data_in) {
		//console.log("create_TSOLO_emu() on 'data'",data_in,data_in.toString().trim());
		var str = data_in.toString().trim(); // trim() chops off trailing \r\n
		if (str == 'ACK') {
			; // ACK is good, but we do not need to respond
		} else if (str == 'huhu') {
			//
			send_orf_hw();
		} else {
			; // all other data.... no response
		}
	});

}


// received broadcast message from target
// .ip: ip address
// .type: G1, G2, S10, R5
function deviceBroadcastRecv( msg, rinfoAddress ) {

	log4.trace( "broadcast from " + rinfoAddress );

	// compatible with first G2 firmware, remove this later
	if ( msg.smtdevtype && !msg.type ) {
		msg.type = msg.smtdevtype;
	}

	// ingore messages from anything except a G2 or SOLO
	if ( msg.type !== "G2" && msg.type !== "SOLO" ) {
		return;
	}

	msg.ip = rinfoAddress;
	msg.is_target = ( msg.type === "G1" || msg.type === "G2" || msg.type === "SOLO" );

	var deviceId = msg.ip; // index on IP address
	if ( msg.type === "SOLO" ) {
		//deviceId += ":" + msg.deviceIP;
		deviceId = msg.serialno;
		log4.trace( "broadcast from SOLO " + msg.deviceIP + " (Server: " + msg.ip + ")" );
	}
	msg.internalId = deviceId;

	var isNew = !data.devices[ deviceId ]; // whether device is new

	log4.trace( "Is new: " + isNew );


	if ( isNew ) {

		// active on network
		msg.is_present = true;

		// allow association to targets
		if ( msg.is_target ) {
			msg.allow_association = true;
		}

		// G2 can only associate frames to itself
		if ( is_G2 ) {
			// DMC 22-Feb-2017: searching for my IP address is not too terribly expensive,
			// typically 6-7ms elapsed time.  Do this every time allows us to respond to
			// our IP address changing (admittedly not very common) and also some odd
			// not-yet-reproduced edge cases out there of us getting "stuck" on a particular
			// self-G2 ip addr.
			var local_ip = undefined;
			// determine if this device is local
			// by searching list of local interfaces for this address
			var nets = require( "os" ).networkInterfaces();
			for ( var i in nets ) {
				var net1 = nets[ i ];
				for ( var j in net1 ) {
					var net2 = net1[ j ];
					if ( net2.address === msg.ip ) {
						log4.debug( "set local ip", msg.ip );
						local_ip = msg.ip;
						break;
					}
				}
			}
			if (msg.ip !== local_ip ) {
				msg.allow_association = false;
			}
		}

		var crud = {};
		crud.devices = {};
		crud.devices[ deviceId ] = msg;
		apply_crud( crud );

	}

	// the actual device object
	var device = data.devices[ deviceId ];

	// if we have a new target, create a listener and try to auto associate it
	if ( isNew && device.allow_association ) { // as a G2, do not create listeners for other G2s
		if ( device.type !== "SOLO" ) {
			frame_create_listener( device );
		}
		for ( var frameId in data.frames ) {
			var frame = data.frames[ frameId ];
			if (
				frame.enabled && !frame.assoc_device_id
				&& frame.assoc_device_id_last === deviceId
			) {
				frame_associate( frameId, deviceId ); // DMC - I think this is a bug; frame_associate is for G1
			}
		}
	}

	// not new and anything important has changed, update it
	// don't do this trivially because it causes a client refresh
	if ( !isNew ) {
		var hasChanged = false;
		if ( !device.is_present ) {
			msg.is_present = true;
			hasChanged = true;
		}
		if ( abs( msg.batteryvalue - device.batteryvalue ) >= 5 ) hasChanged = true;
		if ( msg.shot_valid !== device.shot_valid ) hasChanged = true;
		if ( msg.shot_valid ) {
			if ( msg.shot_x !== device.shot_x ) hasChanged = true;
			if ( msg.shot_y !== device.shot_y ) hasChanged = true;
		}
		if ( msg.name !== device.name ) hasChanged = true;
		if ( msg.type !== device.type ) hasChanged = true;
		if ( msg.remote_rf_level !== device.remote_rf_level ) hasChanged = true;
		if ( msg.ping_health ) {
			device.ping_health = {}; // will be set by apply_crud(), below
			hasChanged = true;
		}
		if ( msg.c2fwversion !== device.c2fwversion) hasChanged = true;
		if ( msg.deviceIP !== undefined && msg.deviceIP !== device.deviceIP ) hasChanged = true;

		if ( 'undefined' != typeof msg.SOLOconnected) {
			if ( msg.type === "SOLO" ) {
				setFrameConnectionState( data.devices[ deviceId ], msg.SOLOconnected ); // propogate to frame(s)
			}
		}
		if ( hasChanged ) {
			var crud = {};
			crud.devices = {};
			crud.devices[ deviceId ] = msg;
			apply_crud( crud );
		}
	}
	// set timer to mark not present after 20 sec
	clearTimeout( device_present_timers[ deviceId ] );
	device_present_timers[ deviceId ] = setTimeout( function() {
		var crud = {};
		crud.devices = {};
		crud.devices[ deviceId ] = { is_present: false };
		apply_crud( crud );
	}, 20000 );

}

var setFrameConnectionState = function( device, isConnected ) { // only used by SOLO for now
	if (!device.assoc_frame_id) return;
	device.assoc_frame_ids_list[device.assoc_frame_id] = {}; // backward-compat

	var crud = undefined; // avoid sending out empty cruds { frames: {} }

	for ( var i in device.assoc_frame_ids_list ) { // iterate over all connected frames
		var frame = data.frames[ i ];
		if ( frame && frame.connected !== isConnected ) {
			if (!crud) crud = { frames: {} };
			crud.frames[ frame._id ] = { connected: isConnected }; // RED/GREEN status
//console.log("setFrameConnectionState connected:",isConnected);
		}
	}
	if (crud) apply_crud( crud );
};
