/*******************************************************************************
 viewer.js: main file for high level client (application level above client.js)

 $Author: dmc $
 $Revision: 2631 $
 $Date: 2019-07-12 14:57:42 -0300 (Fri, 12 Jul 2019) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/public/javascripts/viewer.js $
*******************************************************************************/

/*  high level structure:
 - string() -> frame(), face(), cur_shot(), shots[]
 - frame -> string()
 - shot -> string()
 - cur_string() and cur_shot() may be null
 
*/

var cur_string_id = null;
function cur_string() { // currently active String
	if (cur_string_id == null) return null; // no string selected
	if (g_string_view && cur_string_id == g_string_view.id())
		return g_string_view; // viewing old string
	return g_strings[cur_string_id]; // active string on frame
}

function select_string(id) {
	cur_string_id = id;
	var str = cur_string();
//	maintarget.set_pan(0, 0); // reset view
//	maintarget.zoom = maintarget.zoom_default; // not using set_zoom; rely on maintarget.update later on using this.
	shotbubble.enabled = false; // toggle off shotbubble
	if (str) {
		str.select_last_shot(); // select shot
	}
	//log("select_string (" + ((id == null)?("null"):(str.data.name)) + ")");
}

function frame_is_valid_to_select(frame) {
	if (!frame.data.enabled && !display_admin_mode(true)) return false; // disabled: must be in admin panel
	if (!frame.device() && !admin_check()) return false; // not associated: must be logged in as admin
	if (!frame.data.enabled && !toolbar_system.show_hidden_frames) return false; // disabled, in admin panel, but hidden
	return true;
}

// timer to prevent autoselecting shots as they come in and removing focus
var shot_selection_timer = null;
function set_shot_selection_timer() {
	clearTimeout(shot_selection_timer);
	shot_selection_timer = setTimeout(function() {
		shot_selection_timer = null;
	}, 5000);
}

// return true if we are not trying to view an old string
function live_mode() {
	return g_string_view == null || cur_string() != g_string_view;
}


var printing = false; // set true and update to clean display for printing

// format page and print with window.print
function print_screen() {

	// set printing flag and update screen
	printing = true;
	update();
        
    try {
    // build printing window
    var w = window.open("", "window to print", "width=800px,height=1000px");
    w.document.write("<link rel='stylesheet' href='stylesheets/viewer.css'>");
	
	var div = $("<div>");
	
	// smt logo
	var logo = $("<img src='images/logo.png'>");
	logo.css("width", "300px");
	logo.css("position", "absolute");
	logo.css("left", "0");
	logo.css("top", "0");
	div.append(logo);
	
	// string info area
	var info = $("<div>");
	var str = cur_string();
	var date = new Date();
	if (!live_mode()) {
		info.append(str.data.name + "<br>");
		date = new Date(str.data.ts);
	}
	info.append(date.toDateString() + "<br>");
	info.append(date.toLocaleTimeString());
	
	// current plot scope setting
	if (plotting.enable_view) {
		var scope_x = 0, scope_y = 0;
		if (str.data.shot_params && str.data.shot_params.scope) {
			scope_x = str.data.shot_params.scope_x;
			scope_y = str.data.shot_params.scope_y;
		}
		info.append("<br>Plot windage: " + scope_y);
		info.append("<br>Plot elevation: " + scope_y);
	}

	info.css("font-family", "Helvetica");
	info.css("font-size", "16px");
	info.css("position", "absolute");
	info.css("left", "350px");
	info.css("top", "0");
	div.append(info);

	// main target canvas (fixed size 600x600)
	var target = $("<img src='" + maintarget.canvas[0].toDataURL("image/png") + "'>");
	target.css("position", "absolute");
	target.css("left", "0");
	target.css("top", "100px");
	div.append(target);

	// shot list
	var list = shotlist.div.clone();
	list.css("width", "190px");
	list.css("height", "990px");
	list.css("position", "absolute");
	list.css("left", "600px");
	list.css("top", "0");
	div.append(list);
    
    w.document.write(div[0].outerHTML);
	
	setTimeout(function() { // let the main screen fix itself immediately
        try {
            w.print();
            setTimeout(function() { w.close(); }, 3000);
        } catch (err) {
        }
	}, 1000); // hope it loads in 1 second...
	
	printing = false;
	update();

    } catch (er) {
        //alert(er);
        alert('Your browser doesn\'t support popup windows, try to open the print page in a new tab.');
    }
        
        
}

// entry point: init modules then call init_client()
function init_viewer() {

	tblMainTable = $("#tblMainTable");
	imgServerDown = $("#imgServerDown");

	maintarget.init();
	targetrow.init();
	toolbar.init();
	shotlist.init();
	shotbubble.init();
	plotting.init();

	init_client();
}

// called after all_data or heartbeat message is processed
function data_received(msg, complete) {
	//log("data_received(" + JSON.stringify(msg).length + " bytes)");
	
	// initial data load
	if (complete == true) {

		// select recent user from cookie
		var user_i = get_cookie("recentuser");
		if (g_users[user_i]) {
			log("loading user " + g_users[user_i].data.name + " from cookie");
			user_select(user_i);
		}

		// select recent frame from cookie
		var frame_i = get_cookie("recentframe");
		if (g_frames[frame_i] && g_frames[frame_i].data.enabled) {
			log("loading frame " + g_frames[frame_i].data.name + " from cookie");
			select_string(g_frames[frame_i].string().id());
		}
	}

    if (msg.frames) {
        // if current frame has a new string, select it
        var frame_id = (cur_string())?(cur_string().frame().id()):(null);
        if (frame_id && msg.frames[frame_id] && msg.frames[frame_id].string_id) {
            select_string(msg.frames[frame_id].string_id);
        }
        // check if some string data or state changed for some frame and update targetrow if needed...
        var redrawTargetRow = false;
        for (var i in msg.frames) {
            var frame = msg.frames[i];
            if (frame.string_id !== undefined) {
                targetrow.updateByString(g_strings[frame.string_id]);
            }
            var redrawOnProperties = ['enabled', 'connected', 'assoc_device_id', 'name'];
            if (!redrawTargetRow && _.find(redrawOnProperties, function(p) { return _.has(frame, p); })) {
                // if the frame object have any property of redrawOnProperties then redraw the complete targetrow
                redrawTargetRow = true;
            }
        }
        if (redrawTargetRow) targetrow.redraw();
    }

	// for admin only, if a new frame was just created, select it
	if (display_admin_mode(true) && msg.frames) {
		for (var i in msg.frames) {
			if (msg.frames[i].name == "new") {
				
				// save cookie for this frame index in list
				// need to do this because this is likely a full broadcast data
				set_cookie("recentframe", msg.frames[i]._id);

				select_string(msg.frames[i].string_id);
			}
		}
	}

	// for any string where shots were updated, select last shot
	for (var i in msg.strings) {

		if (!msg.strings[i].shots) continue; // no shots in this message

		// reasons to not auto-select
		if (cur_string() && cur_string().id() == i) {
			
			if (shot_selection_timer != null) continue; // timer to prevent this is active
			
			// none of the shots in this message are valid
			var found = false;
			for (var j in msg.strings[i].shots) {
				var shot = msg.strings[i].shots[j];
				if (shot && shot.valid) {
					found = true;
					break;
				}
			}
			if (!found) continue;

		}

		// okay, do it
		if (g_strings[i]) {
            g_strings[i].select_last_shot();
            targetrow.updateByString(g_strings[i]);
        }
	}
    
	//log("data_received() done");
}

// called anytime anything changes
var demo_allgreen = false;
var demo_invert_connected = false;
function update() {
	// demo_allgreen = true; // uncomment this line to force all targets to "green", for demo purposes.
	// or for testing, can set this in debug console
	if (demo_allgreen) for (f in g_frames) g_frames[f].data.connected = true;
	if (demo_invert_connected) for (f in g_frames) g_frames[f].data.connected = !g_frames[f].data.connected;

	if (!init_data_received) return; // cannot update without any data yet!
	
	if (toolbar.dropdown_open) { // don't update when dropdowns are open
		toolbar.dropdown_need_update = true;
		return;
	}
	
	// trace("update()");
	
	var shotlist_scroll = shotlist.div[0].scrollTop;

	if (g_connected) {
        hide(imgServerDown);
    } else {
        show(imgServerDown);
    }

	// auto select a frame if necessary
	if (live_mode()) { // only if not viewing old string
	
		// try to select a valid string on an enabled frame
		if (!cur_string() || !frame_is_valid_to_select(cur_string().frame())) {
			for (var i in g_frames) {
				if (frame_is_valid_to_select(g_frames[i])) {
					select_string(g_frames[i].string().id());
					break;
				}
			}
		}

		// still no string: select the first frame's string
		if (!cur_string()) select_string(first_value(g_frames).string().id());
		
	}
		
	// delete face target paper canvas if no longer needed
	for (var i in g_faces) {
		g_faces[i].paper_needed = false;
	}
	cur_string().face().paper_needed = true; // needed in case of viewing old string
	for (var i in g_strings) {
		g_strings[i].face().paper_needed = true;
	}
	for (var i in g_faces) {
		if (!g_faces[i].paper_needed) g_faces[i].paper = null;
	}
	
	// allow table to shrink
	maintarget.clear();
	shotlist.clear();
	
	toolbar.update();
	shotlist.update();
	maintarget.update();
	targetrow.update();
	shotbubble.update();
	plotting.update();


	// for ipad, in case table is slightly too large
	$(document).scrollTop(0);

	// always try to restore shotlist scroll position
	shotlist.div.scrollTop(shotlist_scroll);

	//log("update() done");
	
}

// update some things on the string then save it
function save_string(string, name) {
	
	var delta = {};
	
	// user supplied name
	delta.name = name;
	
	// total shots
	delta.shots_mode = g_config.shots_mode;
	
	// shooter count
	delta.shooter_count = string.data.shooter_count || 1;
	
	// save event data if using rstat event
	var event_active = g_config.event_enabled && g_config.event_data.event_id >= 0 && g_config.event_data.match_id >= 0 && g_config.event_data.stage_id >= 0;
	if (event_active) {
		delta.event_data = g_config.event_data;
	}
	
	// record whether the global option is set
	delta.use_role_shooter = g_config.use_role_shooter;
	
	// apply changes locally before computing score string
	update_object(string.data, delta);
	
	delta.score = [];
	for (var i = 0; i < delta.shooter_count; i++) {
		delta.score[i] = calc_string_score(string.data, i);
	}
	
	// apply changes to string before saving
	string.crud(delta);
	
	// save and clear
	g_socket.emit("create_string",
		string.frame().id(),
		string.face().id()
	);
	
}

function update_calibration() {

	// current calibration values	
	var frame = cur_string().frame();
	var prev_x_cal = frame.data.geom.x_cal;
	var prev_y_cal = frame.data.geom.y_cal;

	var shots = cur_string().shots();
	
	var data = []; // for math

	var crud = new_crud(); // build message to send to server
	crud.frames[frame.id()] = {};
	crud.strings[cur_string().id()] = {};
	
	for (var i in shots) {
		var shot = shots[i];
		if (!shot) continue;

		if (shot.measured) {
			data.push({
				sx: shot.x - prev_x_cal, // shot position at zero cal
				sy: shot.y - prev_y_cal,
				mx: shot.measured_x,
				my: shot.measured_y
			});
		}
	}
	
	if (data.length == 0) return;

	var new_x_cal = 0; // for offset
	var new_y_cal = 0;

	var avg_mx = 0, avg_my = 0, avg_sx = 0, avg_sy = 0; // for scale

	for (var i in data) {
		new_x_cal += (data[i].mx - data[i].sx);
		new_y_cal += (data[i].my - data[i].sy);
		avg_mx += data[i].mx;
		avg_my += data[i].my;
		avg_sx += data[i].sx;
		avg_sy += data[i].sy;
	}

	if (data.length > 0) {
		new_x_cal /= data.length;
		new_y_cal /= data.length;
		avg_mx /= data.length;
		avg_my /= data.length;
		avg_sx /= data.length;
		avg_sy /= data.length;
	}
	
	// 2 or more points: compute scale
	if (data.length > 1) {

		// calculate standard deviations for scale
		var std_mx = 0, std_my = 0, std_sx = 0, std_sy = 0;
		for (var i in data) {
			std_mx += (data[i].mx - avg_mx) * (data[i].mx - avg_mx);
			std_my += (data[i].my - avg_my) * (data[i].my - avg_my);
			std_sx += (data[i].sx - avg_sx) * (data[i].sx - avg_sx);
			std_sy += (data[i].sy - avg_sy) * (data[i].sy - avg_sy);
		}
		
		// scale error is the ratio of group size in each direction
		var scale_x = Math.sqrt(std_mx / data.length) / Math.sqrt(std_sx / data.length);
		var scale_y = Math.sqrt(std_my / data.length) / Math.sqrt(std_sy / data.length);
		
		crud.frames[frame.id()].scale = true;
		crud.frames[frame.id()].scale_x = scale_x;
		crud.frames[frame.id()].scale_y = scale_y;
		
	
	} else { // not enough data for scale
		crud.frames[frame.id()].scale = false;
	}

	// apply new offsets
	crud.frames[frame.id()].geom = { x_cal: new_x_cal, y_cal: new_y_cal };
	
	crud.strings[cur_string().id()].shots = [];
	
	// update all existing shots in string
	for (var i in shots) {
		var shot = shots[i];
		if (!shot) continue;

		// new position
		var new_x = shot.x - prev_x_cal + new_x_cal;
		var new_y = shot.y - prev_y_cal + new_y_cal;
        var new_rawX = shot.rawX - prev_x_cal + new_x_cal;
        var new_rawY = shot.rawY - prev_y_cal + new_y_cal;
		
		// update on server
		crud.strings[cur_string().id()].shots[i] = {
			x: new_x,
			y: new_y,
            rawX: new_rawX,
            rawY: new_rawY
		};
		
	}

	// send this request to server
	// do not also update client-side - better to be safe
	g_socket.emit("crud", crud);

	// need to recompute string things
	g_socket.emit("recalc_shot_scores", cur_string().id());
	g_socket.emit("update_string_stats", cur_string().id());

}




