/*******************************************************************************
 viewer_shotbubble.js: client module for shot popup bubble

 $Author: dmc $
 $Revision: 2576 $
 $Date: 2019-05-24 10:43:02 -0300 (Fri, 24 May 2019) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/public/javascripts/viewer_shotbubble.js $
*******************************************************************************/

// calibration measurements:
// client requests server to set measured coords
// server recalibrates and updates shots

// shot call interactions:
// client is updated immediately for display purposes
// client is allowed to get ahead of the server
// client requests update on server

shotbubble = new function() {

	this.init = function() {
		this.div = $("#divShotBubble");
		
		this.imgStrikeShot = $("<img id='imgStrikeShot' src='images/cancel.png'>");
		this.imgAcceptShot = $("<img id='imgAcceptShot' src='images/accept.png'>");
		this.imgSetSighter = $("<img id='imgSetSighter' src='images/sighter.png'>");
		this.imgSetRecord = $("<img id='imgSetRecord' src='images/record.png'>");
		this.imgMissNext = $("<img id='imgSetRecord' src='images/miss-next.png'>");
		this.imgMissPrev = $("<img id='imgSetRecord' src='images/miss-prev.png'>");
	}

	this.clear = function() {
		hide(this.div);
		this.div.empty();
	}

	// for shot call canvas
	this.args = {};
	
	this.canvas = null;
	this.events = null;
	
	this.enabled = false; // whether to show (toggled by shot selection)

	// place at selected row in shotlist
	this.update_position = function() {
		if (!ok(shotlist.focus_tr)) return;
		var top = shotlist.focus_tr[0].offsetTop + shotlist.focus_tr[0].parentElement.parentElement.offsetTop - shotlist.div[0].scrollTop;
		top = Math.min(top, shotlist.div[0].clientHeight - shotbubble.div[0].offsetHeight + 10);
		top = Math.max(top, 0);
		shotbubble.div[0].style.top = top + "px";
	}
	
	this.adjust_shot_scope = function(wind, moa) {
		var shot = cur_string().cur_shot();
		if (!shot) return;
		if (shot.shooter_num < 0) return;
		
		if (!shot.scope) {
			shot.scope = true;
			shot.scope_x = [0, 0, 0];
			shot.scope_y = [0, 0, 0];
		}
		if (wind) shot.scope_x[shot.shooter_num] += moa;
		else shot.scope_y[shot.shooter_num] += moa;
		
		// round both elev and wind to fix switching units
		var round_to = (g_config.option_grid_unit == "mil")?(mil_to_moa(0.1)):(0.25);
		log(round_to);
		shot.scope_x[shot.shooter_num] = round(shot.scope_x[shot.shooter_num] / round_to) * round_to;
		shot.scope_y[shot.shooter_num] = round(shot.scope_y[shot.shooter_num] / round_to) * round_to;
		
		cur_string().crud_shot(cur_string().cur_shot_index, {
			scope: true,
			scope_x: shot.scope_x,
			scope_y: shot.scope_y
		});
	}
	
	this.user_assign_shooter = function(i) {
		var shot = cur_string().cur_shot();
		if (!shot) return;
		
		if (shot.shooter_num != i) {
			cur_string().crud_shot(cur_string().cur_shot_index, {
				shooter_num: i
			});
			g_socket.emit("recalc_shot_scores", cur_string().id());
		}
		shotbubble.enabled = false;
		shotbubble.update();
	}
	
	this.update = function() {
		this.clear();

		if (!this.enabled) return;

		if (!cur_string()) return;
		var shot = cur_string().cur_shot();
		if (!shot) return;

		if (g_config.basic_mode) return; // shotbubble not displayed in basic mode

		//log("shotbubble.update()");
		
		var display = false; // only show if there is content
		this.div.css("text-align", "center");
		
		// show/set scope and hold settings
		if (plotting.enable_view) {
		
			display = true;

			var use_mils = (g_config.option_grid_unit == "mil");

			this.div.append("Sight setting:<br>");

			this.div.append("E: ");
			
			if (plotting.enable_controls) {
				var cmdShotScopeElevDown = $("<button id='cmdShotScopeElevDown' class='title_element'>").text("-");
				cmdShotScopeElevDown.css("width","40px");
				cmdShotScopeElevDown.mousedown(function() {
					shotbubble.adjust_shot_scope(false, (use_mils)?(mil_to_moa(-0.1)):(-0.25));
					maintarget.update_canvas();
					shotbubble.update();
					plotting.update_canvas();
				});
				this.div.append(cmdShotScopeElevDown);
			}

			var elev = 0;
			if (shot.scope && typeof(shot.scope_y) == "object") elev = shot.scope_y[shot.shooter_num];
			else if (shot.scope) elev = shot.scope_y;
			if (use_mils) elev = moa_to_mil(elev);

			this.div.append($("<span class='title_element boldnumber'>").text(elev.toFixed(2)));

			if (plotting.enable_controls) {
				var cmdShotScopeElevUp = $("<button id='cmdShotScopeElevUp' class='title_element'>").text("+");
				cmdShotScopeElevUp.css("width","40px");
				cmdShotScopeElevUp.mousedown(function() {
					shotbubble.adjust_shot_scope(false, (use_mils)?(mil_to_moa(0.1)):(0.25));
					maintarget.update_canvas();
					shotbubble.update();
					plotting.update_canvas();
				});
				this.div.append(cmdShotScopeElevUp);
			}

			this.div.append("<br>");

			this.div.append("W: ");

			if (plotting.enable_controls) {
				var cmdShotScopeWindDown = $("<button id='cmdShotScopeWindDown' class='title_element'>").text("-");
				cmdShotScopeWindDown.css("width","40px");
				cmdShotScopeWindDown.mousedown(function() {
					shotbubble.adjust_shot_scope(true, (use_mils)?(mil_to_moa(-0.1)):(-0.25));
					maintarget.update_canvas();
					shotbubble.update();
					plotting.update_canvas();
				});
				this.div.append(cmdShotScopeWindDown);
			}

			var wind = 0;
			if (shot.scope && typeof(shot.scope_x) == "object") wind = shot.scope_x[shot.shooter_num];
			else if (shot.scope) wind = shot.scope_x;
			if (use_mils) wind = moa_to_mil(wind);
			
			this.div.append($("<span class='title_element boldnumber'>").text(wind.toFixed(2)));

			if (plotting.enable_controls) {
				var cmdShotScopeWindUp = $("<button id='cmdShotScopeWindUp' class='title_element'>").text("+");
				cmdShotScopeWindUp.css("width","40px");
				cmdShotScopeWindUp.mousedown(function() {
					shotbubble.adjust_shot_scope(true, (use_mils)?(mil_to_moa(0.1)):(0.25));
					maintarget.update_canvas();
					shotbubble.update();
					plotting.update_canvas();
				});
				this.div.append(cmdShotScopeWindUp);
			}

			this.div.append("<br>");

			this.div.append("Call: ");
	
			var call_div = $("<div>");
			call_div.css("margin", "0 auto");
			call_div.css("width", "100px");
			call_div.css("height", "100px");
			this.div.append(call_div);
			
			this.draw_shot_call_target(call_div);
			
		}
		
		// ability to reassign shot to a different shooter
		if (live_mode()) {
			
			display = true;
			
			if (user_can_affect_shots() && cur_string().data.shooter_count > 1) {
			 			
				this.div.append(S("STRING_ASSIGN") + "<br>");
				
				var shooter_pos = [];
				if (cur_string().data.shooter_count == 2) shooter_pos = ["L", "R"];
				else shooter_pos = ["L", "M", "R"];
				
				for (var i = 0; i < cur_string().data.shooter_count; i++) {
					var cmdAssignShooter = $("<button id='cmdAssignShooter" + i + "' onmousedown=\"shotbubble.user_assign_shooter(" + i + ")\">");
					cmdAssignShooter.text(shooter_pos[i]);
					cmdAssignShooter.css("font-size", "30px");
					cmdAssignShooter.css("margin", "0px");
					cmdAssignShooter.css("width", "50px");
					cmdAssignShooter.css("margin-bottom", "5px");
					if (shot.shooter_num == i) {
						cmdAssignShooter.addClass("button_green");
					}
					this.div.append(cmdAssignShooter);
				}
				this.div.append("<br>");
			}
//basic_mode

			var insertMissingShot = function(index) {
				g_socket.emit("insert_missed_shot", { stringId: cur_string().id(), shotIndex: index, shooter_num: shot.shooter_num});
				g_socket.emit("update_string_stats", cur_string().id());
			};
			
			this.imgMissPrev.mousedown(function() {
				result = confirm(S("STRING_INSERTMISS_PREV"));
				if (result) {
					insertMissingShot(cur_string().cur_shot_index);
				}
			});
			this.imgMissNext.mousedown(function() {
				result = confirm(S("STRING_INSERTMISS_NEXT"));
				if (result) {
					insertMissingShot(cur_string().cur_shot_index + 1);
				}
			});
			user_can_affect_shots() && this.div.append(this.imgMissNext).append(this.imgMissPrev);
			
			// replay one shot, copy one shot data, paste shot data...
//			if (get_cookie('smtdev') == 'true') {
			if (g_config.is_SMT) {
				// determine frames to allow restore string
				var frames_to_display = [];
				for (var i in g_frames) {
					var frame = g_frames[i];
					if (!frame_is_valid_to_select(frame)) continue; // check against admin/frame status
					frames_to_display.push(i);
				}

				frames_to_display.sort(function(a,b) {
					if (g_frames[a].data.name == g_frames[b].data.name) return 0;
					if (g_frames[a].data.name == "new") return -1;
					if (g_frames[b].data.name == "new") return 1;
					return g_frames[a].data.name.toUpperCase().localeCompare(g_frames[b].data.name.toUpperCase());
				});
		
				var list = [];
				list.push({value: 0, text: "Replay.."});
				for (i in frames_to_display) {
					var id = frames_to_display[i];
					var frame = g_frames[id];
					list.push({value: id, text: frame.data.name});
				}
		
				toolbar.new_dropdown(this.div, list, function(item) {
					var frame = g_frames[item.val()]; // chosen frame to replay onto
					if (!frame) return;
			
					var answer = confirm("CONFIRM: replay on " + frame.data.name+ " ?");
					if (!answer) return;
					g_socket.emit("replay_shot", cur_string().data.shots[cur_string().cur_shot_index], frame.id());

				}, 0);

				// download / paste-in single shot:
				// cancel anytime, even if all shooters have accepted
				toolbar.new_button(this.div, "[COPY]..", function() { // button to copy shotdata to clipboard
					copyTextToClipboard(JSON.stringify(
						cur_string().data.shots[cur_string().cur_shot_index]
						,null,2) // pretty-fied, 2-space indent
					);
				});

				var paste_tgt = toolbar.new_text(this.div, "txt");
				paste_tgt.attr('contenteditable','true').text('[PASTE SHOT HERE]');
				paste_tgt.on("paste",function(e) {
					var pasted =e.originalEvent.clipboardData.getData('text');
					var json_parsed = undefined;
					try {
						json_parsed = JSON.parse(pasted);
					} catch (parse_error) {
						alert("parse error" + parse_error);
					}
					//console.log(json_parsed);
					// console.log("pasted=",e.originalEvent.clipboardData.getData('text'));
					// TODO: process 
					e.preventDefault();
				});

			} // g_config.is_SMT

			user_can_affect_shots() && this.div.append(this.imgStrikeShot);

			if (!shot.strikeout) {
				this.imgStrikeShot.mousedown(function() {
					var shot = cur_string().cur_shot();
					result = confirm(S("STRING_STRIKEOUT"));
					if (result) {
						cur_string().crud_shot(cur_string().cur_shot_index, {
							strikeout: true, strikeout_recalc_autolearn: true,
                            strikeoutTs: new Date().getTime()
						});
						g_socket.emit("update_string_stats", cur_string().id());
						shotbubble.enabled = false;
						shotbubble.update();
					}
				});
				user_can_affect_shots() && this.div.append(this.imgStrikeShot);
			} else {
				this.imgAcceptShot.mousedown(function() {
					var shot = cur_string().cur_shot();
					result = confirm(S("STRING_RECYCLE"));
					if (result) {
						cur_string().crud_shot(cur_string().cur_shot_index, {
							strikeout: false, strikeout_recalc_autolearn: true,
						});
						g_socket.emit("update_string_stats", cur_string().id());
						shotbubble.enabled = false;
						shotbubble.update();
					}
				});
				user_can_affect_shots() && this.div.append(this.imgAcceptShot);
			}
			
			// only can manually mark sighters as admin or match panel
			if (!shot.sighter) {
				this.imgSetSighter.mousedown(function() {
					var shot = cur_string().cur_shot();
					result = confirm(S("STRING_CUTSIGHTER"));
					if (result) {
						var delta_shots = [];
						for (var i = 0; i <= cur_string().cur_shot_index; i++) {
							var s2 = cur_string().data.shots[i];
							if (!s2) continue;
							if (cur_string().data.shooter_count > 1 && s2.shooter_num != shot.shooter_num) continue;
							delta_shots[i] = {sighter: true};
						}
						cur_string().crud({shots: delta_shots});
						g_socket.emit("update_string_stats", cur_string().id());
						shotbubble.enabled = false;
						shotbubble.update();
					}
				});
				user_can_affect_shots() && this.div.append(this.imgSetSighter);
			} else {
				this.imgSetRecord.mousedown(function() {
					var shot = cur_string().cur_shot();
					result = confirm(S("STRING_MARKRECORD"));
					if (result) {
						var delta_shots = [];
						for (var i = cur_string().cur_shot_index; i < cur_string().data.shots.length; i++) {
							var s2 = cur_string().data.shots[i];
							if (!s2) continue;
							if (cur_string().data.shooter_count > 1 && s2.shooter_num != shot.shooter_num) continue;
							delta_shots[i] = {sighter: false};
						}
						cur_string().crud({shots: delta_shots});
						g_socket.emit("update_string_stats", cur_string().id());
						shotbubble.enabled = false;
						shotbubble.update();
					}
				});
				user_can_affect_shots() && this.div.append(this.imgSetRecord);
			}
					
		}
		
		var mic_tap = shot && shot.mics && get_mic_tap_num(shot.mics);
		if (mic_tap) {
			var div_mic_tap = $('<div>').addClass("").text("Mic Tap");
			var div_mic_tap_num = $('<div>').addClass("").text(mic_tap);
			div_mic_tap.append(div_mic_tap_num);
			this.div.append(div_mic_tap);
			display = true;
		} else {
			// warnings/errors?   invalid_reasons star_reasons
			if (shot.valid == false) {
				var div_invalid = $('<div>').addClass("").text("Invalid Shot");
				shot.invalid_reasons && $.each(shot.invalid_reasons, function (idx,item) {
					var div_invalid_reason = $('<div>').addClass("").text(item);
					div_invalid.append(div_invalid_reason);
				});
				this.div.append(div_invalid);
				display = true;
			}
			if (shot.is_starred) {
				var div_starred = $('<div>').addClass("").text("Measurement Warning");
				shot.star_reasons && $.each(shot.star_reasons, function (idx,item) {
					var div_starred_reason = $('<div>').addClass("").text(item);
					div_starred.append(div_starred_reason);
				});
				this.div.append(div_starred);
				display = true;
			}
			// Don't do anything for whack_missing_mics - this will be done (or not) via starred_reasons
		}

		// nothing to put in the bubble
		if (!display) return;
		
		// put at selected row in shotlist
		this.update_position();
		
		//log("shotbubble.update() done");
		show(this.div);
	}

	this.draw_shot_call_target = function(div) {
		log("draw_shot_call_target");

		var zoom_small = 2.1;

		var args = this.args; // for passing parameters around
		args.string = cur_string();
		args.frame = args.string.frame();
		args.face = args.string.face();
		args.shot = args.string.cur_shot();

		// set up display area
		args.w = div.width() - 5;
		args.h = div.height() - 5;
		args.cx = args.w / 2;
		args.cy = args.h / 2;
		args.moa_to_px = Math.min(args.w, args.h) / zoom_small;
		args.mm_to_px = args.face.mm_to_moa * args.moa_to_px;
		
		// rebuild canvas if it does not already exist
		if (ok(this.canvas)) {
			
			if (this.canvas[0].width != args.w || this.canvas[0].height != args.h) {
				this.canvas[0].width = args.w;
				this.canvas[0].height = args.h;
			}
			
		} else {

			this.canvas = $("<canvas>").attr({width: args.w, height: args.h});
			this.events = new Hammer(this.canvas[0]);

			this.events.on("tap", function(e) {
				if (!plotting.enable_controls) return; // not in view only mode

				var args = shotbubble.args;
				
				// move immediately and request update from server
				args.shot.hold = true;
				args.shot.hold_x = (e.center.x - (e.target.parentElement.offsetLeft + shotbubble.div[0].offsetLeft) - args.cx) / args.moa_to_px;
				args.shot.hold_y = -(e.center.y - (e.target.parentElement.offsetTop + shotbubble.div[0].offsetTop) - args.cy) / args.moa_to_px;
				
				args.string.crud_shot(args.string.cur_shot_index, {
    				hold: true,
					hold_x: args.shot.hold_x,
					hold_y: args.shot.hold_y
				});
    			
				maintarget.update_canvas();
				shotbubble.update_canvas();
				plotting.update_canvas();
			});
			
			this.events.get("pan").set({ direction: Hammer.DIRECTION_ALL });

			this.events.on("panstart", function(e) {
				if (!plotting.enable_controls) return; // not in view only mode

				var args = shotbubble.args;

				// move immediately to touch start position
				// keep crosshair fixed offset above touch
				args.shot.hold = true;
				args.shot.hold_x = (e.center.x - (e.target.parentElement.offsetLeft + shotbubble.div[0].offsetLeft) - args.cx - e.deltaX) / args.moa_to_px;
				args.shot.hold_y = -(e.center.y - (e.target.parentElement.offsetTop + shotbubble.div[0].offsetTop) - args.cy - e.deltaY - 30) / args.moa_to_px;
				
				args.pan_last_x = 0;
				args.pan_last_y = 0;
			});

			this.events.on("pan", function(e) {
				if (!plotting.enable_controls) return; // not in view only mode

				var args = shotbubble.args;
				
				var dx = e.deltaX - args.pan_last_x; args.pan_last_x = e.deltaX;
				var dy = e.deltaY - args.pan_last_y; args.pan_last_y = e.deltaY;

				// can't use half-sensitivity because of fixed offset
				args.shot.hold_x += dx / args.moa_to_px;
				args.shot.hold_y -= dy / args.moa_to_px;

				maintarget.update_canvas();
				shotbubble.update_canvas();
				plotting.update_canvas();
			});
			this.events.on("panend", function(e) {
				if (!plotting.enable_controls) return; // not in view only mode

				var args = shotbubble.args;

				args.string.crud_shot(args.string.cur_shot_index, {
    				hold: true,
					hold_x: args.shot.hold_x,
					hold_y: args.shot.hold_y
				});

				maintarget.update_canvas();
				shotbubble.update_canvas();
				plotting.update_canvas();
			});

		}

		div.append(this.canvas);

		this.update_canvas();
	}

	this.update_canvas = function() {
		if (!ok(this.canvas)) return;
		if (!ok(this.canvas[0].parentElement)) return;
		
		log("update_canvas");

		var g = this.canvas[0].getContext("2d");
		var args = this.args;

		g.clearRect(0, 0, args.w, args.h);

		// draw layers in order
		args.face.draw_target_board(g, args);
		args.face.draw_target_image(g, args);
		args.face.draw_target_zones(g, args, draw_fclass_target());

		// draw cross at current call position
		g.beginPath();

		var px = args.cx;
		var py = args.cy;
		
		if (args.shot.hold) {
			px += args.moa_to_px * args.shot.hold_x;
			py -= args.moa_to_px * args.shot.hold_y;
		}

		var cross = 1000;
		g.moveTo(px, py - cross); g.lineTo(px, py + cross);
		g.moveTo(px - cross, py); g.lineTo(px + cross, py);

		g.strokeStyle = "#8F8";
		g.lineWidth = args.moa_to_px * 0.06;
		g.stroke();

	}
}

// https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
function copyTextToClipboard(text) {
  var textArea = document.createElement("textarea");

  //
  // *** This styling is an extra step which is likely not required. ***
  //
  // Why is it here? To ensure:
  // 1. the element is able to have focus and selection.
  // 2. if element was to flash render it has minimal visual impact.
  // 3. less flakyness with selection and copying which **might** occur if
  //    the textarea element is not visible.
  //
  // The likelihood is the element won't even render, not even a flash,
  // so some of these are just precautions. However in IE the element
  // is visible whilst the popup box asking the user for permission for
  // the web page to copy to the clipboard.
  //

  // Place in top-left corner of screen regardless of scroll position.
  textArea.style.position = 'fixed';
  textArea.style.top = 0;
  textArea.style.left = 0;

  // Ensure it has a small width and height. Setting to 1px / 1em
  // doesn't work as this gives a negative w/h on some browsers.
  textArea.style.width = '2em';
  textArea.style.height = '2em';

  // We don't need padding, reducing the size if it does flash render.
  textArea.style.padding = 0;

  // Clean up any borders.
  textArea.style.border = 'none';
  textArea.style.outline = 'none';
  textArea.style.boxShadow = 'none';

  // Avoid flash of white box if rendered for any reason.
  textArea.style.background = 'transparent';


  textArea.value = text;

  document.body.appendChild(textArea);
  textArea.focus();
  textArea.select();

  try {
    var successful = document.execCommand('copy');
    var msg = successful ? 'successful' : 'unsuccessful';
    console.log('Copying text command was ' + msg);
  } catch (err) {
    console.log('Oops, unable to copy');
  }

  document.body.removeChild(textArea);
}


