/*******************************************************************************
 viewer_toolbar.js: client module for top icon and controls toolbar

 $Author: dmc $
 $Revision: 2486 $
 $Date: 2019-01-23 02:35:59 -0400 (Wed, 23 Jan 2019) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/public/javascripts/viewer_toolbar.js $
*******************************************************************************/
var toggled = false;

var toolbar = new function() {

	this.panels = [
		toolbar_user,
		toolbar_system,
		toolbar_info,
		toolbar_arc,
		toolbar_simulator,
		toolbar_match,
		toolbar_string,
		toolbar_plot,
		toolbar_military,
		toolbar_military_tgt,
		toolbar_review
	];

	this.init = function() {
        var that = this;
		this.div = $("#divToolbar");
		this.divWatermarkR = $("#divWatermarkR");
		this.divWatermarkL = $("#divWatermarkL");

		this.back_icon = $("<img src='images/back.png' class='toolbar_icon inline_icon'>");
		this.print_icon = $("<img src='images/print.png' class='toolbar_icon'>");
		this.watermark_iconL = $("<img src='images/watermarkL.png'>").css({
			'max-width': '100%', 'max-height': '100%'
			, 'width': 'auto', 'height': 'auto'
		});
		this.watermark_iconR = $("<img src='images/watermarkR.png'>").css({ 'max-width': '100%', 'max-height': '100%' });;

		// init individual toolbar panels
		for (var i in this.panels) {
			toolbar.panels[i].init();
		}
        
        this._resize();
        $(window).resize(function() {
            that._resize();
        });
	}
	
	this.active_panel = null; // pointer to the currently displayed panel, or null
	
	this.panel_restored = false; // set to true after loading the first time
	
	this.dropdown_open = false; // flag whether dropdown is currently open (to prevent updates)
	this.dropdown_need_update = false; // flag whether we need to update after closing a dropdown
	
	this.back = function() {
		if (!toolbar.active_panel) return true;
		if (toolbar.active_panel.hide()) {
			toolbar.active_panel = null;
			clear_cookie("active_panel");
			return true;
		}
		return false;
	}
	
    this._resize = function () {
        // calculate the max height...
        var maxHeight = $('#tblMainTable').height() - (toolbar.div.offset().top * 2);
        toolbar.div.css('max-height', maxHeight + 'px');
    };
    
	this.set_panel = function(panel) {
		if (panel && panel.show()) {
			toolbar.active_panel = panel;
			set_cookie("active_panel", this.panels.indexOf(panel));
			return true;
		}
		return false;
	}

	this.clear = function() {
		hide(this.div); hide(this.divWatermarkL); hide(this.divWatermarkR);
		this.div.empty();
	}
	
	this.new_button = function(div, text, callback, green) {
		var item = $("<button class='title_element'>");
		item.html(text);
		item.mousedown(callback);
		//new Hammer(item[0]).on("tap", callback);
		if (green) item.addClass("button_green");
		div.append(item);
		return item;
	}
	
	this.new_text = function(div, text) {
		var item = $("<span class='title_element'>")
		item.html(text);
		div.append(item);
		return item;
	}
	
	this.new_checkbox = function(div, text, callback, checked) {
		var item = $("<input type='checkbox'>");
		if (checked) item.prop("checked", true);
		item.change(callback);
		item.bind("touchstart", function(e) { e.originalEvent.allow_touchstart = true; }); // for touch
		this.new_text(div, text);
		div.append(item);
		return item;
	}

	// remember, val() returns a string and "0" is true
	this.new_dropdown = function(div, options, callback, defaultvalue, param) {
		var item = $("<select class='title_element'>");
		
		// wrap in a div which can be positioned
		var inner = $("<div>");
		inner.append(item);
		div.append(inner);
		inner.css("display", "inline-block");
		inner.css("position", "relative");
		inner.css("vertical-align", "middle");
		inner.css("padding", "5px");

		for (i in options) {
			var str = "<option value='" + options[i].value + "'>" + options[i].text + "</option>";
			if (i == -1) item.prepend(str);
			else item.append(str);
		}
		item.val(defaultvalue);

		item.selectmenu(); // apply jquery-ui
		
		// set up scrolling
		var widget = item.selectmenu().selectmenu("menuWidget");
		widget.css("height", "300px");
		widget.bind("touchstart", function(e) { e.originalEvent.allow_touchstart = true; });
		widget.bind("touchmove", function(e) { e.originalEvent.allow_touchmove = true; });

		if (callback) item.on("selectmenuchange", function() {
			setTimeout(callback, 10, item, param);
		}); // need timeout to avoid UI bugs
		
		// prevent updates while open
		item.on("selectmenuopen", function() {
			toolbar.dropdown_open = true;
		});
		item.on("selectmenuclose", function() {
			toolbar.dropdown_open = false;
			if (toolbar.dropdown_need_update) {
				toolbar.dropdown_need_update = false;
				update();
			}
		});
		
		log(item);
		
		return item;
	}

	this.update = function() {
		this.clear();

		if (toolbar_military.is_mil()) {
			$("#divToolbar").addClass('mil_background'); // adjust background colour
			$("#divTargetRow").addClass('mil_background');
		}

		toolbar.dropdown_open = false;

		//log("toolbar.update()");

		// check if should restore panel from cookie
		if (!this.panel_restored) {
			this.panel_restored = true;
			var prev_panel = this.panels[get_cookie("active_panel")];
			if (prev_panel == toolbar_user && !cur_user()) prev_panel = null; // don't load user panel with no user logged in
			if (prev_panel != toolbar_review && !this.set_panel(prev_panel)) {
				clear_cookie("active_panel");
			}
		}
		
		// if a panel is active
		if (this.active_panel != null) {
			
			// draw the back button
			this.back_icon.mousedown(function() {
				if (toolbar.back()) update();
			});
			this.div.append(this.back_icon);
			
			this.active_panel.draw(this.div);
			
		// icon toolbar
		} else {
			
			// draw icons
			for (var i in this.panels) {
				if (this.panels[i].ready()) {
					this.panels[i].icon[0].arg = i;
					this.panels[i].icon.mousedown(function() {
						if (toolbar.set_panel(toolbar.panels[this.arg])) update();
					});
					this.div.append(this.panels[i].icon);
				}
			}

			if (window.print && !toolbar_military.is_mil()) {
				this.print_icon.mousedown(function () {
					print_screen();
				});
				this.div.append(this.print_icon);
			}


			if (cur_user()) {
				this.div.append("<br>");
				var str = "Logged in as " + user_display_name(cur_user().id());
				if (ro_check()) str += " (" + toolbar_military.stringname_RangeOfficer() + ")";
				this.div.append($("<span class='title_element'>").text(str)); // todo
				
				var reg = user_is_registered(cur_user().id());
				if (!reg) {
					this.div.append("<br>");
					var item = toolbar.new_text(this.div, "You are not registered for this event!");
					item.addClass("text_red");
				} else if (typeof(reg) == "object") {
					this.div.append("<br>");
					var item = toolbar.new_text(this.div, "Registered with class " + reg.guntype);
					item.addClass("text_green");
				}

			}

			if (toolbar_system.simulate_mode) {
				this.div.append("<br>");
				this.div.append($("<span class='title_element'>").text("*** Shot Simulator is ON ***")); // todo
			}

		}

		// DMC -- a better info display?
		if (toolbar_system.info_plus) {
			var info_div = $('<div id=infodiv>'); // viewer_shotlist.js "info" button will fill this in
			this.div.append(info_div);
			info_div.css("white-space", "pre"); //  so that newline and spaces are shown
			info_div.css("overflow-y", "scroll");
			info_div.css("max-height",maintarget.div.height()-$('#divToolbar').height() - $('#divTargetRow').height());
		}
		
		this.div.css("max-height", maintarget.div.height() - $('#divTargetRow').height());

		show(this.div);
		if (toolbar_military.is_mil()) {
			this.divWatermarkL.css('bottom', $('#divTargetRow').height())
				// .css('height', ($('#divMainTarget').height() - $('#divTargetRow').height()) / 2)
				// .css('width', $('#divMainTarget').width() / 3)
				.css('height', '10%')
				.css('width', '20%')
				.css('opacity', 0.4)
				.css('left', '10px')
				.append(this.watermark_iconL)
				;
			this.divWatermarkR.css('bottom', $('#divTargetRow').height())
				//			.css('height', ($('#divMainTarget').height() - $('#divTargetRow').height()) / 2)
				.css('width', $('#divMainTarget').width() / 3)
				.css('height', '15%')
				.css('width', '20%')
				.css('right', '10px')
				.css('opacity', 0.4)
				.append(this.watermark_iconR)
				;
			show(this.divWatermarkL);
			show(this.divWatermarkR);
		}
		//log("toolbar.update() done");
	}

}


