/*******************************************************************************
 server.js: main server file

 $Author: dmc $
 $Revision: 2766 $
 $Date: 2020-06-29 17:04:28 -0300 (Mon, 29 Jun 2020) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/server.js $
 $Id: server.js 2766 2020-06-29 20:04:28Z dmc $
*******************************************************************************/

var log4js = require('log4js');
var _log = log4js.getLogger('server.js');

// recent shot for broadcasting
global.recent_shot_valid = false;
global.recent_shot_x = 0;
global.recent_shot_y = 0;

// store the local ip (for comparison against broadcasts received)
global.local_ip = 0;

global.ip_temperature = -100; // temperature from probe on network

// includes
var fs = require("fs"); // file system
var exec = require("child_process").exec;
eval(fs.readFileSync("./smt.common.lib/public/javascripts/utils.js")+""); // common utils
eval(fs.readFileSync("./smt.common.lib/public/javascripts/math.js")+""); // common utils
eval(fs.readFileSync("./smt.common.lib/server_devices.js")+""); // handle broadcasts from any device
eval(fs.readFileSync("./smt.common.lib/server_frames.js")+""); // connection to targets
eval(fs.readFileSync("./server_storage.js")+""); // file system storage method
eval(fs.readFileSync("./server_rstat.js")+""); // rstat database

var archive_collection = require('./smt.common.lib/models/collection.archive.js');

// the grand kluge-o-rama, until we properly modularize things:
//global.storage_read_string = storage_read_string;
//global.storage_import_targetfaces = storage_import_targetfaces;
global.storage_write_data = storage_write_data;
global.storage_write_faces = storage_write_faces;
global.storage_write_all = storage_write_all;
global.frameid_emit_shot_direct = frameid_emit_shot_direct;
global.namedframe_calc_timings_sim_shot = namedframe_calc_timings_sim_shot;
global.update_object = update_object;


var ClientStats = require('./lib/utils/client-stats.js');
var clientStats = new ClientStats();

var lobots_control = require("./lobots_control.js"); // high level LOBOTS control functions
var mic_mapping = require('./smt.common.lib/mic_mapping.js');
var serverjs_handlers = require('./smt.common.lib/serverjs_handlers.js');
var new_storagejs = require('./smt.common.lib/new_storage.js');
var common_serverjs = require('./smt.common.lib/common.server.js');

// set time automatically only once per session
var time_has_been_set = false;



function server_main() { // a thousand+ line ugly disaster, but there's a reason for it...

 console.log("starting server_main()")




//storage_init(); // get ready to read files, create default data files if necessary






global.data = new_storagejs.storageReadDataSync(); // load active data from files (frames, faces, active strings)
global.archive = new_storagejs.archive2; // dynamically generated
global.archive_collection = archive_collection; // load string archive (inactive strings without shots)
archive_collection.init(global.archive); // load string archive (inactive strings without shots)
storage_cleanup(); // remove any unwanted files
storage_update_diskspace(); // set diskspace

// logging off by default, turn it on with env. var.
data.config.enable_logging = !!process.env['SMT_ENABLE_LOGGING'] || false;
data.config.version_string = g_version_string;
data.config.other_G2_button = !!process.env['SMT_MANUAL_OTHER_TARGETS'] || false;

var string_stats_timers = {}; // for more efficient calling of string stats


log("frames: " + numberof(data.frames));
log("faces: " + numberof(data.faces));
log("strings: " + numberof(data.strings));
log("users: " + numberof(data.users));
log("events: " + numberof(data.events));
log("archive: " + numberof(archive));

// reset frame connection/association status
frames_init();

// ready to talk to devices
devices_init();

// ready to talk to clients
var express = require("express"); // start express
var app = express();
app.use(express.static("public")); // set directory
app.use(express.static("smt.common.lib/public")); // set directory
app.use(express.bodyParser()); // for uploading files
// connect.multipart() will be removed in connect 3.0
// visit https://github.com/senchalabs/connect/wiki/Connect-3.0 for alternatives
// connect.limit() will be removed in connect 3.0

var listening_port = process.env['SMT_LISTEN_PORT'] || 4000; // likely to be 80 for G2
var server = app.listen(listening_port); // start server
global.io = require("socket.io")(server); // attach socket.io

// global except handler to prevent crashing
if (0) process.on("uncaughtException", function (err) {
	console.error("!!! *** UNHANDLED ERROR *** !!!");
	console.error(err);
	console.log("Avoiding a crash but you are living on the edge now...");
});

// Lodi: IP temperature probe #1..... Tycon tpdin-monitor-web
var ip_temperature_frequency = 60000;
function update_temperature_1() {
	http = require("http");
	http.get("http://192.168.0.21/status.xml", function(response) {
	    var body = "";
	    response.on("data", function (chunk) {
	        body += chunk;
	    });
	    response.on("end", function () {
	        var start = body.indexOf("<temp2>");
	        var end = body.indexOf("</temp2>");
			var temp = Number(body.substring(start+7, end));
			if (start >= 0 && end >= 0 && !isNaN(temp) && temp > -18) {
				global.ip_temperature = temp;
				log("set temperature from IP: " + temp);
			} else {
				log("parsed invalid temperature: " + temp);
//                console.log('DMC-TE 1 error reading temprature:', temp);
			}
			setTimeout(update_temperature_1, ip_temperature_frequency);
		});
        response.on('error', function(err) { /*console.log('DMC-TE 2 error reading temprature:', err);*/ });
	}).on("error", function (err) {
		log("http error reading temperature!");
//        console.log('DMC-TE 3 error reading temprature:', err);
		setTimeout(update_temperature_1, ip_temperature_frequency);
	});
};

// IP temperature probe 'bisley'..... a specially-modified G3 at .0.100
var ip_temperature_frequency = 60000;
function update_temperature_bisley() {
	http = require("http");
	http.get("http://192.168.0.100:3000/temperature.xml", function(response) {
	    var body = "";
	    response.on("data", function (chunk) {
	        body += chunk;
	    });
	    response.on("end", function () {
	        var start = body.indexOf("<temp2>");
	        var end = body.indexOf("</temp2>");
			var temp = Number(body.substring(start+7, end));
			if (start >= 0 && end >= 0 && !isNaN(temp) && temp > -18) {
				global.ip_temperature = temp;
				log("set temperature from IP: " + temp);
			} else {
				log("parsed invalid temperature: " + temp);
//                console.log('DMC-TE 1 error reading temprature:', temp);
			}
			setTimeout(update_temperature_bisley, ip_temperature_frequency);
		});
        response.on('error', function(err) { /*console.log('DMC-TE 2 error reading temprature:', err);*/ });
	}).on("error", function (err) {
		log("http error reading temperature!");
//        console.log('DMC-TE 3 error reading temprature:', err);
		setTimeout(update_temperature_bisley, ip_temperature_frequency);
	});
};
update_temperature_1(); // initial call, to kick things off
update_temperature_bisley(); // initial call, to kick things off


// temperature for S7 and S25 (but not G2)
var Servertemperature_poll_rate = 10000;
delete global.Servertemperature;
setInterval(function () {
	exec("cat /sys/class/thermal/thermal_zone0/temp",function(err,stdout,stderr){
		// output is degrees-C times 1000... e.g. 52500
		if (!err && typeof stdout == 'string'){
			crud = { config: { Servertemperature: {} } }; // build crud message to apply change to active data
			crud.config.Servertemperature = Number(stdout)/1000;
			apply_crud(crud); // apply crud message
		}
	});
}, Servertemperature_poll_rate);

// new client connection
// setup message handlers
io.on("connection", function(socket) {
    _log.debug('connection', socket.id);
    clientStats.tick(socket.id, Date.now(), undefined, socket.request.connection.remoteAddress);
	//console.log("DMC io-connection socket.request.connection=",socket.request.connection.remoteAddress);
	//console.log("DMC io-connection Object.keys(socket)=",Object.keys(socket));
    
	// client requests list of string involving a user
	socket.on("get_archive", function(user_id, event_id, final, recent, only_framename) {
		
		var start = performanceLog('start get_archive...');
		var ts_limit = Date.now() - 1000 * 60 * 60 * 24 * 14; // within two weeks from now

		var list = [];
		for (i in archive) {
			var str = archive[i];

			if (recent && str.ts < ts_limit) continue;

			if (event_id >= 0 && (!str.event_data || str.event_data.event_id != event_id)) continue;
			
			if (final && str.shots_mode > 0) {
				var include = false;
				for (j in str.score) {
					if (str.score[j].count == str.shots_mode) {
						include = true;
						break;
					}
				}
				if (!include) continue;
			}

			if (user_id && str.use_role_shooter) {
				var include = false;
				if (str.role_scorer == user_id) {
					include = true;
				}
				for (j in str.role_shooter) {
					if (str.role_shooter[j] == user_id) {
						include = true;
						break;
					}
				}
				if (!include) continue;
			}
			
			if ("" != only_framename) { // filter only one targetframe
				var frame = data.frames[str.frame_id];
				if (frame && frame.name != only_framename) continue;
			}

			list.push(str);
		}
		
		log("sending archive to client (" + list.length + ")");
		performanceLog('get_archive... list generated... start to send...', start);
		socket.emit("archive", list);
		performanceLog('finished sending archive to client... start stringify...', start);
		var bytes=JSON.stringify(list).length;
		performanceLog('finished sending archive to client...  ' + bytes + ' bytes', start);
	});

	serverjs_handlers.attach_socketio_handlers(socket,io); // "the future"....

	// client requests a new string on a frame/face
	socket.on("create_string", serverjs_handlers.create_string);
	
	socket.on("revive_string", function(str_id, frame_id) {
		log("reviving string " + str_id + " on frame " + data.frames[frame_id].name);
		var str_old = data.strings[data.frames[frame_id].string_id];
		new_storagejs.storage_read_string_async(str_id,function(err,str_new){
			if (err) {
				log("error opening string!");
			} else {
				delete data.strings[str_old._id]; // remove existing string
				str_new._id = new_storagejs.storage_get_unique_id(); // assign a new id to avoid conflict
				data.strings[str_new._id] = str_new; // put new string in main data
				data.frames[frame_id].string_id = str_new._id; // frame uses new string
				str_new.frame_id = frame_id; // string uses this frame
				broadcast_data(); // update everyone
			}
		});
	});
	
	// replay_string now in serverjs_handlers

	// requests to associate targets, enable/disable frames, and simulate a shot
	socket.on("frame_set_enabled", frame_set_enabled);

	socket.on("frame_associate", frame_associate);
	socket.on("frame_associate_G2", frame_associate_G2);
	socket.on("frame_unassociate", frame_unassociate);

	socket.on("frame_simulate_shot", frame_simulate_shot);
	
	socket.on("frame_delete", frame_delete);
	socket.on("frame_create", frame_create);

	socket.on("lobots_updown", lobots_control.lobots_updown);
	socket.on("lobots_set_red_light", lobots_control.set_red_light);
	socket.on("lobots_set_green_light", lobots_control.set_green_light);
	socket.on("lobots_set_illumination_light", lobots_control.set_illumination_light);

	socket.on("face_delete", face_delete);
	
	socket.on("recalc_shot_scores", recalc_shot_scores);
	socket.on("update_string_stats", update_string_stats);

	socket.on("shutdown", function() {
		log("shutting down...");
		storage_write_all(function() {
//            if (global.is_G2 || global.is_R5 || global.is_S3 || global.is_S50) {
            if (0 == process.getuid()) {
                exec("sync");
                // ask MCU to turn off power (message order_set_shut_down 0xF0)
                if ('function' == typeof global.shutdown_power) {
                    global.shutdown_power(); // this gets queued, but it takes a short while
                    // for the message to reach the MCU and the MCU to then turn off the power.
                }

                // we'll get here, but we should wait for the MCU power shutdown command to run....
                setTimeout(function() {
                    exec("poweroff; sleep 10; poweroff -f");
                }, 2000); // if the MCU poweroff command does not work in 2s, shut down via linux

            } else { // S10 trimslice, etc
                exec("sudo -n poweroff");
            }
            // just for windows since poweroff has no effect
            setTimeout(function() {process.exit(); }, 3000) ; // wait 3s, to give time for above commands to complete
        });
	});
	
	socket.on("stop_process", function() {
		log("closing process...");
		storage_write_all(function() {
            setTimeout(function() {process.exit(0); }, 500);
        });
	});
	
	socket.on("reboot_server", function() {
		log("rebooting...");
		storage_write_all(function() {
            setTimeout(function() {
                exec("sync; sleep 2; reboot -f");
            }, 500);
        });
	});

	
	socket.on("clear_old_strings", function(days) {
		storage_clear_old_strings(days);
	});

	// set system time to this exact value
	function reset_time(ts) {
		_log.debug("resetting time...", ts);

		// thanks to: http://stackoverflow.com/questions/10645994/node-js-how-to-format-a-date-string-in-utc
		var new_date = new Date(ts).toISOString()
			.replace(/T/, ' ') // replace T with a space
			.replace(/\..+/, '');
		// "new_date" is now in UTC
		var cmd1 = "date --utc --set='" + new_date + "'";
		var cmd2 = "sudo -n " + cmd1;
		var cmd = cmd1 + ' || ' + cmd2;
		console.log("running command",cmd);
		exec(cmd);
		
		socket.emit("update");
	}
	
	// client requests to set time from pressing button
	socket.on("reset_time_now", reset_time);
	
	// client sends the current time, set to this once only
	socket.on("reset_time_load", function(ts) {
		if (!time_has_been_set) {
			reset_time(ts);
			time_has_been_set = true;
		}
	});
	
	socket.on("write_tcp", function(ipaddr,port, val) { // for JRPL
		write_tcp(ipaddr,port,val);
	});

	socket.on("client_return_ping", function(msg) { // DMC
		msg.recv_ts = Date.now();
		msg.round_trip = msg.recv_ts - msg.ts;
		// somehow we need to log this data
		
        clientStats.tick(socket.id, msg.ts, msg.round_trip, socket.request.connection.remoteAddress);
        
//console.
log("DMC client_return_ping","id=",socket.id
,"rtt=",msg.round_trip
//,msg
//,"ip=",socket.request.connection.remoteAddress // works at first but sooner or later this becomes "undefined"
);

//https://stackoverflow.com/questions/6563885/socket-io-how-do-i-get-a-list-of-connected-sockets-clients
//console.log("DMC2 #clients=",io.engine.clientsCount,Object.keys(io.engine.clients));

	});
		
	socket.on("set_remote_log", function(i, val) {
		var sock = io.sockets.sockets[i];
		sock.remote_log_enabled = val;
		sock.emit("set_remote_log", val);
	});
	
	socket.on("remote_log", function(msg) {
		log("[CLIENT] " + msg);
	});

	socket.on("backup_all_data", function() {
		storage_write_data_immediate(data, function(err) {
			// error or not, we are going to proceed...
			new_storagejs.backup_all_data(socket);
		});
	});

	socket.on("flash_mcu_firmware", function() {
		log("re-flash MCU firmware ...");
		storage_write_all(function() {
            exec("cd updata && cp G2_mcu_image.bin mcu_image.bin" // get mcu firmware file ready
                + " && echo -en '\x0f\x0f\x0f'  > /dev/ttyATH0" // reset MCU
                + " && echo 0 > /sys/class/gpio/gpio15/value" // pull down UART enable line
                + " && echo re-flashed MCU firmware `date` >> /stamp.txt" // pull down UART enable line
                + " && reboot");
        });
	});

	// remove user and broadcast
	socket.on("user_delete", function(user_id) {
		log("user_delete", user_id);
		delete data.users[user_id];
		log("now there are " + numberof(data.users));
		broadcast_data();
	});
	
	// rstat
	socket.on("rstat_sync", function() {
		rstat_sync(socket);
	});
	socket.on("rstat_string_action", function(string_id, action) {
		var str = archive[string_id];
		if (str) str.rstat_string_action = action;
	});
	
	socket.on("event_delete", function(event_id) {
		delete data.events[event_id];
		data.config.event_enabled = false;
		broadcast_data();
	});
	
	socket.on("event_delete", function(event_id) {
		delete data.events[event_id];
		data.config.event_enabled = false;
		broadcast_data();
	});
	
	socket.on("generate_results", function(event_id, match_id, stage_id) {
		var results = generate_results(event_id, match_id, stage_id);
		socket.emit("results", results);
	});
	

});

// handlers






app.get('/firmware-copy', function(req, res) {
	var execSync = require("child_process").execSync;
	var stdout_result = execSync("sh ./firmware-copy.sh");
    res.send(stdout_result || 'ERROR, empty stdout');
    res.end();
});

app.get('/ping', function(req, res) {
    var ip = req.headers['x-forwarded-for'] || req.connection.remoteAddress;
    // DMC 2017-09-24: sometimes 'ip' is not set:
    ip = ('string' == typeof ip) && ip.substring(ip.lastIndexOf('.') + 1, ip.length);
    ip && res.send(ip);
    res.end();
});

app.get('/client-stats-data', function(req, res) {
    var activeConnections = [];
    for(item in io.sockets.sockets) {
        activeConnections.push(io.sockets.sockets[item].id);
    }
    res.json(clientStats.getSummaryClients(activeConnections));
});

// must be called after all of the above, because we include a handler for "*", which must be last
serverjs_handlers.attach_app_handlers(app); // "the future"....

// autosave every 1 min
// save data, not archive, because archive is already saved when it changes
setInterval(function() {
	if (global.is_G2) return; // too dangerous to auto-write out data on a G2, so don't do it.
	if (!data_changed) return; // only actually do the write if data has changed
console.log("server.js interval call to storage_write_data()");
	storage_write_data(data); data_changed = false;
}, 1000*60);


console.log("server ready -- $Revision: 2766 $");

} // function server_main()


// function recv_shot(shot, frame_id) { // moved to smt.common.lib/common_server.js

// permanently delete a frame
function face_delete(face_id) {
	log("face_delete", face_id);

	// cannot delete the last face
	if (numberof(data.faces) == 1) return;
	
	// find a different face to assign to strings that use this one
	var face_id_fallback = null;
	for (var i in data.faces) {
		if (i != face_id) {
			face_id_fallback = i;
			break;
		}
	}
	if (!face_id_fallback) return;
	
	// change active strings that use this face
	for (var i in data.strings) {
		var str = data.strings[i];
		if (str.face_id == face_id) {
			str.face_id = face_id_fallback;
		}
	}
	
	// delete archived strings that used this face
	for (var i in archive) {
		var str = archive[i];
		if (str.face_id == face_id) {
			delete archive[i];
		}
	}
	
	// for image targets, remove the image file:
	var name = data.faces[face_id] && data.faces[face_id].image && data.faces[face_id].image.src;
	if (name) {
		var name_full = "public/"+name;
		//console.log("DMC: face_delete, found fn=",name_full,'exists=',fs.existsSync(name_full)); //,'lstatsync=',fs.lstatSync(name_full));
		rm_symlinkSync(name_full); // remove targetface image file
	}

	// remove from data
	delete data.faces[face_id];

	// force full update to clients
	broadcast_data();

}

//function face_import(face_data) // migrated to new_storage.js

// calculate shot circle parameters for a string
function update_string_stats(string_id) {
	log("update_string_stats", string_id);
	var start = performanceLog('start update_string_stats...');

	var str = data.strings[string_id];
	if (!str) return;
	
	var face = data.faces[str.face_id];
	if (!face) return;
	
	var crud = {};
	crud.strings = {};
	crud.strings[str._id] = {};
	// do autolearning-things...
	common_serverjs.update_string_stats_autolearn(string_id);
	performanceLog('...update_string_stats autolearn...', start);
	
    var nShooter = str.shooter_count || 1;

    update_string_stats_shooter(crud, str, undefined);
    
    if (nShooter > 1) {
        for (n = 0; n < nShooter; n++) {
            update_string_stats_shooter(crud, str, n);
        }
    }

	apply_crud(crud);
	performanceLog('...update_string_stats done', start);

}
global.update_string_stats = update_string_stats; // klugey; TODO: merge with SOLO/G2 and move to a module

function update_string_stats_shooter(crud, str, shooterId) {
	var start = performanceLog('start update_string_stats_shooter...');
    // collect list of included xy shot positions
	var pos = [];
	var score = [0];
	var x_min = 100000, x_max = -100000;
	var y_min = 100000, y_max = -100000;
	    
	for (var i in str.shots) {
		var s = str.shots[i];
		if (!s.valid || s.strikeout || s.sighter) continue;
        
        if (shooterId !== undefined && s.shooter_num != shooterId) continue;
		
		// add shot to calculation list
		pos.push({x: s.x, y: s.y});
		
		// update min/max
		x_min = min(x_min, s.x);
		x_max = max(x_max, s.x);
		y_min = min(y_min, s.y);
		y_max = max(y_max, s.y);
	}
	
	// find group size and smallest circle which encloses all points
	if (pos.length >= 2) {
		
		// group center: average
		var group_center = {x: 0, y: 0};
		for (var i in pos) {
			group_center.x += pos[i].x;
			group_center.y += pos[i].y;
		}
		group_center.x /= pos.length;
		group_center.y /= pos.length;

		// group size: maximum distance between two shots
		var max_distance = 0;
		for (var i in pos) {
			for (var j in pos) {
				// distance squared, for performance
				var dist = (pos[i].x - pos[j].x) * (pos[i].x - pos[j].x) + (pos[i].y - pos[j].y) * (pos[i].y - pos[j].y);
				if (dist > max_distance) max_distance = dist;
			}
		}
		max_distance = sqrt(max_distance); // dist = square root

		function dist_between(a, b) {
			return sqrt((a.x - b.x)*(a.x - b.x) + (a.y - b.y)*(a.y - b.y));
		}

		// define the circle of 2 or 3 points
		function circle_of_points(c) {
			if (c.length != 2 && c.length != 3) {
				return null;
				
			// circle of two points
			} else if (c.length == 2) {
				return {
					center: {
						x: (c[0].x + c[1].x) / 2,
						y: (c[0].y + c[1].y) / 2
					},
					radius: dist_between(c[0], c[1]) / 2
				};
				
			// circle of three points
			} else if (c.length == 3) {

				var e10 = {x: c[1].x - c[0].x, y: c[1].y - c[0].y};
				var e20 = {x: c[2].x - c[0].x, y: c[2].y - c[0].y};

				var A = [[e10.x, e10.y], [e20.x, e20.y]];
				var b = [(e10.x*e10.x + e10.y*e10.y) / 2,
						 (e20.x*e20.x + e20.y*e20.y) / 2];
				
				var det = A[0][0]*A[1][1] - A[0][1]*A[1][0];
				
				if (abs(det) > 0) {
					var cx = (A[1][1]*b[0] - A[0][1]*b[1])/det;
					var cy = (A[0][0]*b[1] - A[1][0]*b[0])/det;
					return {
						center: {
							x: c[0].x + cx,
							y: c[0].y + cy
						},
						radius: sqrt(cx*cx + cy*cy)
					};
				} else {
					return null;
				}
				
			}
					
		}

		// recursive algorithm for smallest enclosing circle found here:
		// http://www.sunshine2k.de/coding/java/Welzl/Welzl.html
		// https://webdocs.cs.ualberta.ca/~contest/code/OLD/src/min_circle.c
		// !! P and C are passed by reference so make sure they are not modified
		function recursive_circle(P, C) {
			if (P.length == 0 || C.length == 3) {
				return circle_of_points(C);
			}
			var P_star = P.slice();
			var p_i = Math.floor(rnd() * P.length); // random index in P
			var p = P_star.splice(p_i, 1)[0]; // select and remove from P
			var B = recursive_circle(P_star, C);
			if (B == null || dist_between(p, B.center) > B.radius) {
				var C_new = C.slice();
				C_new.push(p);
				return recursive_circle(P_star, C_new);
			}
			return B;
		}
		
		var result = recursive_circle(pos, []);
	performanceLog('...update_string_stats_shooter done', start);

        if (shooterId === undefined && result) {
            crud.strings[str._id].use_string_stats = true;
            crud.strings[str._id].stats_circle_center = result.center;
            crud.strings[str._id].stats_circle_radius = result.radius;
            crud.strings[str._id].stats_width = x_max - x_min;
            crud.strings[str._id].stats_height = y_max - y_min;
            crud.strings[str._id].stats_group_size = max_distance;
            crud.strings[str._id].stats_group_center = group_center;
        } else if (result) {
            if (crud.strings[str._id]['shooterStats'] === undefined) {
                crud.strings[str._id]['shooterStats'] = [];
            }
            crud.strings[str._id]['shooterStats'][shooterId] = {
                use_string_stats: true,
                stats_circle_center: result.center,
                stats_circle_radius: result.radius,
                stats_width: x_max - x_min,
                stats_height: y_max - y_min,
                stats_group_size: max_distance,
                stats_group_center: group_center
            }
        }
	} else {
        if (crud.strings[str._id]['shooterStats'] === undefined) {
            crud.strings[str._id]['shooterStats'] = [];
        }
		crud.strings[str._id]['shooterStats'][shooterId] = { use_string_stats: false };
	}
    
}

// recalculate shot scores if face changes
function recalc_shot_scores(string_id) {

	log("recalc_shot_scores", string_id);
	
	var str = data.strings[string_id];
	if (!str) return;

	var face = data.faces[str.face_id];
	if (!face) return;
	
	var crud = {};
	crud.strings = {};
	crud.strings[str._id] = { shots: [] };
	
	for (var i in str.shots) {
		var shot = str.shots[i];
		if (!shot) continue;

        if (shot.rawX === undefined) shot.rawX = shot.x;
        if (shot.rawY === undefined) shot.rawY = shot.y;
        
        shot.x = shot.rawX;
        shot.y = shot.rawY;
        
        if (face.parallax_offset_x) shot.x = face.parallax_offset_x + shot.rawX;
        if (face.parallax_offset_y) shot.y = face.parallax_offset_y + shot.rawY;
        
        if (face.parallax_offset_x || face.parallax_offset_y) {
//            console.log("DMC: parallax correction applied", face.parallax_offset_x, face.parallax_offset_y, shot.x, shot.y);
        }

		// handle switching from 2 shooters to 1
		var shooter_num = shot.shooter_num;
		if (shooter_num >= str.shooter_count) shooter_num = 0;
	
		var fclass = face.scoringzones_f && str.fclass_target && str.fclass_target[shooter_num];
	
		// check if bullet diameter needs to change
		// changes to the bullet size are not crud'ed to the client so will not take effect in view until refresh
		// if the display ever uses this value, this should be fixed
		if (data.config.use_role_shooter && str.role_shooter && str.role_shooter[shooter_num]) {
			var user = data.users[str.role_shooter[shooter_num]];
			if (user && user.bullet_dia) {
				shot.bullet_dia = user.bullet_dia;
			}
		} else shot.bullet_dia = 0; // back to default

		var z = common_serverjs.calc_scorezone(shot, face, fclass);
	
		crud.strings[str._id].shots[i] = {
			value: z.name,
			score: z.score
		};
        
        if (shot.valid) {
            crud.strings[str._id].shots[i].x = shot.x;
            crud.strings[str._id].shots[i].y = shot.y;
        }
	}
	
	apply_crud(crud);

}


function generate_results(event_id, match_id, stage_id) {
	
	// cumulative score by user
	var results = {};
	
	for (i in archive) {
		var str = archive[i];
		
		// must have users defined
		if (!str.use_role_shooter || !str.role_shooter) continue;
		
		// must be from the selected event/match/stage
		if (!str.event_data || str.event_data.event_id != event_id) continue;
		if (match_id >= 0 && str.event_data.match_id != match_id) continue;
		if (stage_id >= 0 && str.event_data.stage_id != stage_id) continue;
		
		// for each score (multiple shooters)
		for (j in str.score) {
			var user_id = str.role_shooter[j];
			if (!data.users[user_id]) continue;
			var user_score = results[user_id] || 0;
			user_score += str.score[j].rstat_score;
			results[user_id] = user_score;
		}
		
	}
	
	return results;
	
}

// auto delete shots we don't care about
setInterval(function() {
	
	var crud = {};
	crud.strings = {};
	
	var found = false;

	// DMC 14-July-2017: allow autopurge to be turned off
	if (!data.config.autopurge_invalid_shots) return; // do we auto-purge?

	for (var i in data.strings) {
		var str = data.strings[i];
		
		// ignore short strings
		// note that this counts valid, invalid, and previously deleted shots
		if (!str.shots || str.shots.length < 100) continue;

		crud.strings[i] = { shots: [] };
		
		for (var j in str.shots) {
			var shot = str.shots[j];
			if (!shot) continue;
			
			// age in minutes
			var age = (Date.now() - shot.ts) / 60000;
			
			// reverse index (0 is most recent)
			var index = str.shots.length - j - 1;
			
			// invalid and older than an hour
			var remove = !shot.valid && age > 60;
			
			// old valid, but more than 100 shots ago, and older than 16 hours
			if (!remove) remove = index > 100 && age > 60*16;
			
			if (remove) {
				crud.strings[i].shots[j] = 0;
				found = true;
			}
			
		}
		
	}
	
	if (found) apply_crud(crud);
	
}, 10*60*1000); // 10 mins

var cached_ip_ports = {};
function write_tcp(ipaddr,port,val) {  // for JRPL

	var ip_port = ipaddr + ':' + port;
	log('write_tcp(),',ip_port,val);
	var sock = cached_ip_ports[ip_port];
	if (!sock) {
		var net = require('net');
		sock = new net.Socket();
		sock.connect(port,ipaddr,function(e,x) {
//			console.log('DMC: socket connected, e=',e,'x=',x);
		});
		sock.on('error',function(ex) { // e.g. ECONNREFUSED
			// just quietly ignore.  delete socket now, so that a future
			// call will re-attempt the connection
			delete cached_ip_ports[ip_port];

		});
		cached_ip_ports[ip_port] = sock;
	}
	var sock = cached_ip_ports[ip_port];
	if (sock) sock.write(val);
}


//g_version_string = "S50-181110-DMCTEST4 $Revision: 2766 $"; //
//g_version_string = "S50-181202-TEST-2351"; //start merge/migrate SOLO thing - server_main(), send_delta()
//g_version_string = "S50-181202-TEST-2364"; // server_main() bugfix; SOLO health
//g_version_string = "S50-181202-TEST-2368"; // apply software updates
//g_version_string = "S50-181221-TEST-"; // new_storage merging
//g_version_string = "S50-190105-TEST2-2437"; // fa-server logging
//g_version_string = "S50-190105-TEST3-2438"; // always save invalid shots
//g_version_string = "S50-190107-2439"; // initial S7 image
//g_version_string = "S50-190115-2472"; // Freeform mode target viewer
// shot delay for non-F-Class; Freeform mode tweaks;
// cleanup menus; get rid of micmapping_perframe
//g_version_string = "S50-190116 $Revision: 2766 $";
//g_version_string = "S50-190117 $Revision: 2766 $"; // bugifx create_image; patch-003
//g_version_string = "S50-190123 $Revision: 2766 $"; // mil skinning; patch-004
//g_version_string = "S50-190304 DMC-WORK $Revision: 2766 $"; // DMC work... patch-005
//g_version_string = "S50-190305 TEST-006 $Revision: 2766 $"; // testing... patch-006... events pass-through
//g_version_string = "Server-190309 PATCH-007 $Revision: 2766 $"; // ... patch-007...  bugfix draw_zero_div, recent_shot - change lightblue emphasis to red disc
//g_version_string = "Server-190318 PATCH-008 $Revision: 2766 $"; // ... patch-008... merge hearbeats/crud/delta into serverjs_handlers
//g_version_string = "Server-190328-TEST $Revision: 2766 $"; // ... GET /archive ; delete object /import-json
//g_version_string = "Server-190403-TEST $Revision: 2766 $"; // ... TSOLO ORF/ACK
//g_version_string = "Server-190520 PATCH-009 $Revision: 2766 $"; // ... bugfix attempt for "strings do not have valid frames"
//g_version_string = "Server-190524-TEST-001 $Revision: 2766 $"; // ... incremental.. shotbubble display shot error/warnings
//g_version_string = "Server-190524-TEST-002 $Revision: 2766 $"; // ... incremental.. data backup/recovery + client.js Refuse to fail.  "Please report this to SMT or your dealer\n"
//g_version_string = "Server-190524-PATCH-010"; // shotbubble display shot error/warnings.. data backup/recovery--- + client.js Refuse to fail.  "Please report this to SMT or your dealer\n"
//g_version_string = "Server-190529-PATCH-011"; // bugfix .toFixed() crash common.server.js; revive_string
//g_version_string = "Server-190531-PATCH-012"; // sanity-checking on msg.strings, to avoid "1 string(s) do not have a valid face!" etc
//g_version_string = "Server-190613-PATCH-013"; // misc. cleanup, further "strings do not have valid frames/faces" fixes, TSOLO reset
//g_version_string = "Server-190621-hotfix"; // Lodi hotfixes - fix reject_shot bug; speed up calc_shot_valid()
//g_version_string = "Server-190709-PATCH-014"; // patch w/Lodi hotfixes
//g_version_string = "Server-190712-PATCH-015"; // disable zoom/pan reset on changing frames; shotbubble auto-show on "*" only for ADMIN/RO
//g_version_string = "Server-190724-test016-1"; // new_storage deprecate archive.txt; fairly major rewrite
//g_version_string = "Server-190809-PATCH-017"; // bugfix pairs/threes shots not showing
//g_version_string = "Server-190809-PATCH-018"; // bugfix insert miss
//g_version_string = "Server-190909-PATCH-019"; // feature - update TSOLO-C2 software
//g_version_string = "Server-190924-PATCH-020"; // hide wifi health unless /health_on, TSOLO-C2 fw update unless /fwtweaks_on, RSTAT unless /rstat_on; reset C2 on save unless /c2noresetsave_on; fix SHUTDOWN for S7
//g_version_string = "Server-191008-PATCH-021"; // improved TSOLO-C2 server.js logic
//g_version_string = "Server-191009-PATCH-022"; // ADMIN user editor add "reset password" button
//g_version_string = "Server-191123-PATCH-023-TEST"; // show C2FW version; improved C2 server.js (timeout & close socket)
//g_version_string = "Server-191124-PATCH-023-TEST2"; // for FSOLO(007) to work - disable timeout on inbound traffic
//g_version_string = "Server-200117-PATCH-023C-TEST3"; // ghost/missing shot fix - angle-screening
//g_version_string = "Server-20-0519"; // rstat.txt file for rstat string save statuses
//g_version_string = "200528-PATCH-024"; // /set_altnetwork
//g_version_string = "200620-PATCH-025.001"; // S25 & S7 temperature reading from /sys/class/thermal/thermal_zone0/temp
//g_version_string = "200625-PATCH-025.002"; // bugfix/defensive-programmging for match/events selection (Norfolk popup)
g_version_string = "200625-PATCH-025.003"; // RSTAT cleanup; new_storage.js handle ".txt" string files; use native ZLIB

new_storagejs.new_storage_init(server_main); // get ready to read files, create default data files if necessary; callback and start main server loop
