/*eslint camelcase: "error"*/

var EventEmitter    = require( "events" ).EventEmitter;
var util            = require( "util" );
var log             = require( "log4js" ).getLogger( "OrfFAHandler" );
var ShotFA          = require( "./shot-fa.js" );

function OrfFAHandler() {
    // this.numberOfDataBytes = 32; // v.1 we expect this: ORF-0345-1145,1129,1104,1130 -62
    this.numberOfDataBytes = 64; // v.2 we expect this: ORF-0345-1145,1129,1104,1130,-62
    // sw vers=101 we expect this: ORF-0345-1145,1129,1104,1130,21,-62
    // ... first packet can be: ORF-0345-1145,1129,1104,1130,0,,0
    this.name = "OrfFA";
    // this.orfRegexp = /ORF-([^-]+)-(\d+),(\d+),(\d+),(\d+) (-{0,1}\d{1,2})/; // old
    this.orfRegexp = /ORF-([^-]*)-(\d+),(\d+),(\d+),(\d+),{0,1} *(\d*),{0,1}(-{0,1}\d{0,2})\r*\n{0,1}/; // old+new
    // this.orfRegexp = /ORF-([^-]+)-(\d+),(\d+),(\d+),(\d+),(-{0,1}\d+)\r*\n/; // new
    this.orfRegexp_huhu = /ORF-HW:(\d+)-FW:(\d+)\r*\n{0,1}/; // "huhu" gives a different kind of ORF - "ORF-HW:100-FW:101\r\n"

    // we expect ORFs every 5s
    this.orf_timeout_value = 6000; // if nonzero, send "ORF-TIMEOUT" if no ORF within this time
    this.orf_timeout_timer = undefined;
}

util.inherits( OrfFAHandler, EventEmitter );

/**
 * 
 * RAW DATA IN 192.168.0.127:44439: 4f52462d 3030303334352d 343039352c313232342c343039352c31323333202d3734
 * 
 *  [TRACE] fa-server.js - data received ascii=ORF-000345-4095,1224,4095,1233 -74
bb bb bb bb bb dc // framing


offset  len sample data
------  --- -----------
0       4   4f 52 46 2d // ORF-
4       30 30 30 33 34 35 2d // 000345-
11      34 30 39 35 2c // "4095,"
16      31 32 32 34 2c // "1224,"
21      34 30 39 35 2c // "4095,"
26      31 32 33 33 20// "1233 "
31      2d3734 // " -74" signal strength



 */


OrfFAHandler.prototype.onMessage = function( data/* array of byte values */, idByte ) {
    var data_str = Buffer( data ).toString();
    //log.debug( "OrfFAHandler data_str=",data_str );
    var old_way = false;
    if (old_way) {
        var lens = [
            // offset  sample data
            // ------  -----------
            4, // 0       4f 52 46 2d // ORF-
            5, // 4       30 30 30 33 34 35 2d // 0345-
            5, // 11      34 30 39 35 2c // "4095,"
            5, // 16      31 32 32 34 2c // "1224,"
            5, // 21      34 30 39 35 2c // "4095,"
            5, // 26      31 32 33 33 20// "1233 "
            4  // 31      2d 37 34 // " -74" signal strength
        ];
        var offsets = [];
        var i;
        for (offsets[0] = 0, i = 1; i < lens.length; i++) {
            offsets[i] = offsets[i - 1] + lens[i - 1];
        }

        function str_get(str, fld_no, offsets, length) {
            return data_str.slice(offsets[fld_no], offsets[fld_no] + length[fld_no]);
        }


        // log.debug( "OrfFAHandle 0",str_get(data_str,0,offsets,lens) );
        // log.debug( "OrfFAHandle 1",str_get(data_str,1,offsets,lens) );
        // log.debug( "OrfFAHandle 2",str_get(data_str,2,offsets,lens) );
        // log.debug( "OrfFAHandle 3",str_get(data_str,3,offsets,lens) );
        // log.debug( "OrfFAHandle 4",str_get(data_str,4,offsets,lens) );
        // log.debug( "OrfFAHandle 5",str_get(data_str,5,offsets,lens) );
        // log.debug( "OrfFAHandle 6",str_get(data_str,6,offsets,lens) );
        /*
        shot.temperature58 = this._calcTemperature( buf.readUInt16BE( 2 + temperaturesOffsetSOLO + 0*4 ) );
        shot.temperature47 = this._calcTemperature( buf.readUInt16BE( 2 + temperaturesOffsetSOLO + 1*4 ) );
        shot.temperature26 = this._calcTemperature( buf.readUInt16BE( 2 + temperaturesOffsetSOLO + 2*4 ) );
        shot.temperature13 = this._calcTemperature( buf.readUInt16BE( 2 + temperaturesOffsetSOLO + 3*4 ) );
        */

        var orf_msg = {
            // orf:str_get(data_str,0,offsets,lens), // "ORF-"
            serialno: str_get(data_str, 1, offsets, lens).slice(0, -1), // chop off trailing "-"
            temprdg_58: parseInt(str_get(data_str, 2, offsets, lens)), // handles trailing ","
            temprdg_47: parseInt(str_get(data_str, 3, offsets, lens)), // handles trailing ","
            temprdg_26: parseInt(str_get(data_str, 4, offsets, lens)), // handles trailing ","
            temprdg_13: parseInt(str_get(data_str, 5, offsets, lens)), // handles trailing " "
            remote_rf_level: parseInt(str_get(data_str, 6, offsets, lens)),
        };

        var SOLOtemperatures = {
            temperature58: ShotFA.prototype._calcTemperature(orf_msg.temprdg_58),
            temperature47: ShotFA.prototype._calcTemperature(orf_msg.temprdg_47),
            temperature26: ShotFA.prototype._calcTemperature(orf_msg.temprdg_26),
            temperature13: ShotFA.prototype._calcTemperature(orf_msg.temprdg_13)
        }
        ShotFA.prototype._calcMeanTemperature(SOLOtemperatures); // calc shot.SOLOtemperatures.mean_temp
        orf_msg.mean_temp_uncal = SOLOtemperatures.mean_temp; // uncalibrated
    } else { // new way, using regexp
        var re_match = data_str.match(this.orfRegexp);
        if (!re_match) { // should never happen, but just in ase...
            return;
        }
        /*
        Fri May  4 16:36:57 2018 daemon.info sh[19144]:   input: 'ORF-0345-4095,4095,1226,4095 -63' ]
        Fri May  4 16:36:57 2018 daemon.info sh[19144]: DMC, orf-regexp [ 'ORF-0345-4095,4095,1226,4095 -63',
        Fri May  4 16:36:57 2018 daemon.info sh[19144]:   '0345',
        Fri May  4 16:36:57 2018 daemon.info sh[19144]:   '4095',
        Fri May  4 16:36:57 2018 daemon.info sh[19144]:   '4095',
        Fri May  4 16:36:57 2018 daemon.info sh[19144]:   '1226',
        Fri May  4 16:36:57 2018 daemon.info sh[19144]:   '4095',
        Fri May  4 16:36:57 2018 daemon.info sh[19144]:   '-63',
        Fri May  4 16:36:57 2018 daemon.info sh[19144]:   index: 0,
        Fri May  4 16:36:57 2018 daemon.info sh[19144]:   input: 'ORF-0345-4095,4095,1226,4095 -63' ] numberOfDataBytes 64
        */
        this.orf_msg = {
            serialno: re_match[1],
            temprdg_58: parseInt(re_match[2]), // handles trailing ","
            temprdg_47: parseInt(re_match[3]), // handles trailing ","
            temprdg_26: parseInt(re_match[4]), // handles trailing ","
            temprdg_13: parseInt(re_match[5]), // handles trailing " " or ","
            next_shot_no: parseInt(re_match[6]),
            remote_rf_level: parseInt(re_match[7]),
        };
        var SOLOtemperatures = {
            temperature58: ShotFA.prototype._calcTemperature(this.orf_msg.temprdg_58),
            temperature47: ShotFA.prototype._calcTemperature(this.orf_msg.temprdg_47),
            temperature26: ShotFA.prototype._calcTemperature(this.orf_msg.temprdg_26),
            temperature13: ShotFA.prototype._calcTemperature(this.orf_msg.temprdg_13)
        }
        ShotFA.prototype._calcMeanTemperature(SOLOtemperatures); // calc shot.SOLOtemperatures.mean_temp
        this.orf_msg.mean_temp_uncal = SOLOtemperatures.mean_temp; // uncalibrated
    }

    // console.log("ORF data = ",this.orf_msg,SOLOtemperatures);
    // console.log("ORF!", data_str, this.orf_msg.temprdg_13);
    // console.log("ORF!", data_str,".",this.orf_msg.remote_rf_level);
    //if (isNaN(this.orf_msg.next_shot_no)) console.log("NaN!!!",re_match[0],"data_str(",data_str.length,")=",data_str);
    this.emit( "ORF", this.orf_msg );
    this.arm_orf_timeout()
};

OrfFAHandler.prototype.arm_orf_timeout = function( ) {

    // clearTimeout(this.orf_timeout_timer);
    // if (this.orf_timeout_value) {
    //     var orf_timeout_function = (function (self,msg) {
    //         return function(msg) {
    //             self.emit("ORF-TIMEOUT", msg);
    //         }
    //     })(this,orf_msg); // closure, bind current "orf_msg"
    //     this.orf_timeout_timer = setTimeout(orf_timeout_function,this.orf_timeout_value);
    //     // console.log("arm orf-time @ ",this.orf_timeout_value)
    // }
    clearTimeout(this.orf_timeout_timer);
    if (this.orf_timeout_value) {
        var orf_timeout_function = (function (self,msg) {
            return function(msg) {
                self.emit("ORF-TIMEOUT", msg);
            }
        })(this,this.orf_msg); // closure, bind current "orf_msg"
        //this.orf_timeout_timer = setTimeout(orf_timeout_function,this.orf_timeout_value + 6000); // red 12s after last transmission - 28 May 2018
        // DMC 12-June-2018, orf_timeout_timer is dynamically calculated to be 120% of actual orf interval, so
        // use that to send the ORF-TIMEOUT
        this.orf_timeout_timer = setTimeout(orf_timeout_function,this.orf_timeout_value); // red 20% after missing ORF
        // console.log("OrfFAHandler.prototype.arm_orf_timeout @ ",this.orf_timeout_value)
    }
}


module.exports = OrfFAHandler;
