/*******************************************************************************
 utils.js: misc utils, for both client-side browser js and server-side nodejs use

 $Author: dmc $
 $Revision: 2665 $
 $Date: 2019-08-09 15:36:25 -0300 (Fri, 09 Aug 2019) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smt.common.lib/public/javascripts/utils.js $
*******************************************************************************/

// get a (relatively) unique ID, used for new objects
var id_counter = 0; // todo: move this to file
function get_unique_id() {
	return Date.now() + id_counter++;
}

// http://stackoverflow.com/questions/5538972/console-log-apply-not-working-in-ie9
if (Function.prototype.bind && console && typeof console.log == "object") {
    // Note: I.E. 8 does not have Array "forEach" method!
    $.each(
    [
      "log" //,"info","warn","error","assert","dir","clear","profile","profileEnd"
    ],
    function (idx,method) {
        console[method] = this.bind(console[method], console);
    }, Function.prototype.call);
}

function log() { // usage: logdebug(level, ....) or logdebug (....)
	if (typeof(g_config) != "undefined" && g_config && g_config.enable_logging == false) return;
	if (typeof(data) != "undefined" && data && data.config && data.config.enable_logging == false) return;
	var aa = Array.prototype.slice.call(arguments);
	//aa.unshift("[LOG]"); // prepend
	console.log.apply(console, aa); // 'console' needed here
	
	//console.log(new Error().stack); // for tracking down the source
	
	// server
	if (typeof(data) != "undefined" && typeof(io) != "undefined" && io) {
		io.emit("log", aa);
	}
	
	// client remote logging
	if (typeof(g_socket) != "undefined" && typeof(remote_logging) != "undefined" && remote_logging) {
		g_socket.emit("remote_log", aa);
	}
}

// console.trace, client side only
function trace() {
	if (typeof(g_config) != "undefined" && g_config && g_config.enable_logging == false) return;
	if (typeof(data) != "undefined" && data && data.config && data.config.enable_logging == false) return;
	var aa = Array.prototype.slice.call(arguments);
	//aa.unshift("[LOG]"); // prepend
	console.trace.apply(console, aa); // 'console' needed here
	
	// server
	if (typeof(data) != "undefined" && typeof(io) != "undefined" && io) {
		io.emit("log", aa);
	}
	
	// client remote logging
	if (typeof(g_socket) != "undefined" && typeof(remote_logging) != "undefined" && remote_logging) {
		g_socket.emit("remote_log", aa);
	}
}

function ok(object) {
	return typeof(object) != "undefined" && object != null;
}

// select first (random) element in associative array
function first_key(array) {
	for (key in array) return key;
	return null;
}

function first_value(array) {
	for (key in array) return array[key];
	return null;
}

function numberof(arr) {
	if (!arr) return 0;
	return Object.keys(arr).length;
}

//
// Problem: > JSON.stringify([,,undefined,,4]) --> '[null,null,null,null,4]'
// -- we can't distinguish between undefined and null
//
//

// for network, to merge delta into object
// cannot really handle deletion because the crud message will delete the field from the heartbeat
// (undefined == null) => true
// null should be ignored. when setting elements of an array, all lower elements will be null
// setting a field to undefined will not propagate across network
// to clear a field, set to 0, so this will work: "if (!field)"
function update_object_original(obj, data) {
	for (var key in data) {
		var type = typeof(data[key]);
		if (type == "object" && data[key] != null) { // source is a real object
			if (typeof(obj[key]) != "object" || obj[key] == null) { // destination is not a real object
				if (Array.isArray(data[key])) obj[key] = []; // new array
				else obj[key] = {}; // new associative array
			}
			update_object(obj[key], data[key]); // recursive
		} else if (data[key] != null) { // source is scalar
			obj[key] = data[key]; // set
		}
	}
}

g_special_delete_obj = { _delete: true }; // globally available

function test_object_delete(datum) { // is 'datum' a special "please delete this" object?
	var datum_type = typeof datum; // WARNING: typeof null == 'object'
	if (datum_type == "object" && null != datum && datum._delete === true && Object.keys(datum).length == 1) {
		//console.log("test_object_delete -- detected special _delete object");
		return true;
	}
	return false;
}

function update_object(obj, data, opts /*object; optional */) {
	// opts.record_dirty /*bool*/ -- if true, then set opts._dirty == true if any changes to "obj" happen
	// opts.create_delta /*bool*/ -- true if we are creating a delta
	var create_delta = opts && opts.create_delta; // create a delta for sending
	var record_dirty = opts && opts.record_dirty;
	var trace_dirty = opts && opts.trace_dirty;
	for (var key in data) {
		var datum = data[key]; // if datum == null, we DO NOT apply anything, obj[key] is unchanged
		var datum_type = typeof datum; // NB typeof null == 'object'
		if (datum_type == "object") { // source is a real object
			if (datum == null) { // datum == null, and incidentally type == "object"
				// we can't delete an object using "null" as the changed-data, because we
				// have to handle the case [,,,x] which is stringified into [null,null,null,x]
				//
				// So, we *IGNORE* all deltas with newvalues of null
				//
			} else if (test_object_delete(datum) && !create_delta) {
				// special case to handle an actual delete
				// (but, if we are creating a delta, we must propagate special
				// object rather than acting upon it)
				//console.log("update_object - delete object obj[",key,"]=",obj[key]);
				if (record_dirty) opts._dirty = true;
				if (trace_dirty) console.log("update_object _dirty delete object, key=",key);
				delete obj[key]; // actually delete -- on target object
			} else { // recursive call to process an assignment of an object value
				if (typeof(obj[key]) != "object" || obj[key] == null) { // destination is not a real object
					if (record_dirty) opts._dirty = true;
					if (trace_dirty) console.log("update_object _dirty dest!='object', key=",key);
					if (Array.isArray(datum)) obj[key] = []; // new array
					else obj[key] = {}; // new associative array
				}
				update_object(obj[key], datum, opts); // recursive
			}
		} else { // not object --> scalar
			if (record_dirty && obj[key] != datum) {
				if (trace_dirty) console.log("update_object _dirty scalar",obj[key],'!=',datum);
				opts._dirty = true;
			}
			obj[key] = datum; // set
		}
	}
}

// show/hide elements with css
function hide(div) {
	if (!div || !div[0] || !div[0].style) return;
	div[0].style.visibility = "hidden";
}

function show(div) {
	if (!div || !div[0] || !div[0].style) return;
	div[0].style.visibility = "visible";
}

// set css top, bottom, left, right, width, height
function set_pos(div, dir, pos) {
	if (!div || !div[0] || !div[0].style) return;
	div[0].style[dir] = pos + "px";
}

function accuratetime() {
	var res = process.hrtime();
	log(res);
	return res[0] + res[1] * 1e-9;
}

// whether main target (cur_string) should be f-class or not
function draw_fclass_target_test(curString) {

    // target face has no f-class zones defined
    if (!curString.face().data.scoringzones_f) return false;

    // string has option set
    if (curString.data.fclass_target && curString.data.fclass_target[cur_shooter]) {
        return true;
    }

    // default
    return false;
}

function isShotHiddenByDelay(shot, string) {
    var now = (new Date()).getTime() - cur_ts_local_error,
        fclass = draw_fclass_target_test(string);

    if (g_config.shot_display_delay 
        && live_mode() 
        && shot.valid // no delay on displaying incoming crossfires (which are !shot.valid)
        && (fclass || !g_config.shot_display_delay_FClass_only)
        && now < shot.ts + g_config.shot_display_delay * 1000 
        && now > shot.ts - 1000
    ) {
        return true;
    }
    return false;
}

function isShotNearInTimeToCurShot(shot, string, within_time /*ms*/) {
	within_time = within_time || 3000;
	var cur_shot = string.cur_shot();
	if (!cur_shot || !shot) return false;
	
	if (Math.abs(cur_shot.ts - shot.ts) < within_time) {
		return true;
	}
	
	return false;
}

// simple hash of given string
// https://stackoverflow.com/a/7616484
String.prototype.hashCode = function() {
	var hash = 0, i, chr;
	if (this.length === 0) return hash;
	for (i = 0; i < this.length; i++) {
	  chr   = this.charCodeAt(i);
	  hash  = ((hash << 5) - hash) + chr;
	  hash |= 0; // Convert to 32bit integer
	}
	return hash;
};

function convertTemperatureTxt( tC ) {
	// in case tC is a string:
	tC = +tC;
	if ( g_config !== undefined && g_config.temperature !== undefined ) {
		switch( g_config.temperature ) {
			case "degF":
				var tF = tC * 1.8 + 32;
				return (tF.toFixed(0)) + "F";
//				return (tF.toFixed(0)) + " °F";
		}
	}
	return (tC.toFixed(0)) + "C";
//	return (tC.toFixed(0)) + " °C";
}

// compute score for one shooter only (excludes sighters)
// operates directly on string.data, so it can be used generally
function calc_string_score(string_data, shooter_i) {

	var total_score = [0];
	var count = 0;
	var shooter_count = string_data.shooter_count || 1;
	
	var speeds = [], th=[], ph=[], ycoords=[];

	var hideByDelay = false,
		update_again_soon = -1,
		now = (new Date()).getTime() - cur_ts_local_error;

	var shots = string_data.shots;
	for (var i in shots) {
		var shot = shots[i];

		if (!shot || !shot.valid || shot.sighter || shot.strikeout || shot.source == 'missed_shot') continue;
		if (shooter_count > 1 && shot.shooter_num != shooter_i) continue;

        // F-Class 7-second delay....
        // Quick'n'dirty: use cur_string, because string_data don't contains function frame()
		if (isShotHiddenByDelay(shot, cur_string())) {
			hideByDelay = true;
			var show_shot_delay = shot.ts + g_config.shot_display_delay * 1000 - now; // how long until we un-hide?
			if (show_shot_delay < update_again_soon) update_again_soon = show_shot_delay;
//console.log("DMC debug hideByDelay",hideByDelay,"update_again_soon",update_again_soon,"show_shot_delay",show_shot_delay);
			continue;
		}

		count++;

		// add score columns
		if (shot.score) {
			if (Array.isArray(shot.score)) {
				for (var i in shot.score) {
					if (!total_score[i]) total_score[i] = 0;
					total_score[i] += shot.score[i];
				}
			} else {
				total_score[0] += shot.score;
			}
		}
		
		speeds.push(shot.v);
		ycoords.push(shot.y);
		if (shot.res && undefined != shot.res.theta) th.push(shot.res.theta);
		if (shot.res && undefined != shot.res.phi) ph.push(shot.res.phi);

	}
	var s_speeds = calc_mean_sd(speeds); // calc speed stats
	var s_ycoords = calc_mean_sd(ycoords); // calc ycoords stats

	var ss = calc_mean_sd(th); // calc theta stats
	var th_avg = ss.mean, th_sd = ss.sd;
	
	var ss = calc_mean_sd(ph); // calc phi stats
	var ph_avg = ss.mean, ph_sd = ss.sd;

	// calculate total score string
	var score_string = "";

	var face = g_faces[string_data.face_id];
	if (face && face.data.options && face.data.options.shot_format) {
		score_string = face.data.options.shot_format;
		for (var i in total_score) {
			score_string = score_string.replace("{" + (Number(i)+1) + "}", total_score[i]);
		}
		for (var i = 0; i < 10; i++) {
			score_string = score_string.replace("{" + (i+1) + "}", 0);
		}
	} else {
		for (var i in total_score) {
			score_string += total_score[i] + ".";
		}
		score_string = score_string.substring(0, score_string.length-1);
	}

	var rstat_score = total_score[0];
	if (total_score[1]) rstat_score += total_score[1] * 0.001;

	return {string: score_string, count: count, rstat_score: rstat_score
		,v_avg: s_speeds.mean, v_sd: s_speeds.sd, v_stderr: s_speeds.stderr, v_min: s_speeds.min, v_max: s_speeds.max
		,v_mr: s_ycoords.meanrange
		,th_avg:th_avg, th_sd:th_sd
		,ph_avg:ph_avg, ph_sd:ph_sd
		,hideByDelay: hideByDelay // for 7-second delay
	};
}

function calc_mean_sd(samples) {
	if (samples.length < 1) return {mean: 0, sd: 0, min:0, max:0, stderr:0, meanrange: 0 };
	var _avg = 0;
	var _sd = 0;
	var _mr = 0;
	var _min, _max;
	_min = _max = samples[0];
	for (i in samples) {
		_avg += samples[i];
		if (samples[i] < _min) _min = samples[i];
		if (samples[i] > _max) _max = samples[i];
	}
	_avg /= samples.length;
	for (i in samples) {
		_sd += (samples[i] - _avg)*(samples[i] - _avg);
		_mr += Math.abs(samples[i] - _avg);
	}
	_sd = Math.sqrt(_sd / samples.length);
	_mr = _mr / samples.length;
	return {mean:_avg, sd: _sd, min: _min, max: _max, stderr: _sd/Math.sqrt(samples.length),  meanrange:_mr };
}

function get_mic_tap_num(mics) { // return mic #1..8 if just one mic tapped, otherwise 0
	// has one mic only
	var mic_tap = 0;
	for (var i in mics) {
		if (mics[i] >= 0) {
			if (mic_tap) { // already have a mic
				mic_tap = 0;
				break;
			}
			mic_tap = Number( i ) + 1;
		}
	}
	return mic_tap;
}

